from xml.dom.core import *

class Walker:

	def walk(self, root):
		if root.NodeType == DOCUMENT:
			assert root.documentElement.NodeType == ELEMENT
			return self.walk1(root.documentElement)
		else:
			return self.walk1(root)
		
	def walk1(self, node):
		if node.NodeType == ELEMENT:
			self.startElement(node)
			for child in node.getChildren():
				self.walk1(child)
			self.endElement(node)
		else:
			self.doText(node)





"""
This is a convenience XML utility library for those who want to
make XML software in Python.

It's currently not released, because it needs much more work, but it's
included in the xmlproc package because xmlproc uses it.

Version 0.01 - 12.Mar.98
   Lars Marius Garshol - larsga@ifi.uio.no
   http://www.stud.ifi.uio.no/~larsga/download/python/xml/
"""
   
import string,re

class XML_catalog:
    """An SGML Open catalog file handler
    
    Only handles a simple syntax subset. Each entry must be on separate
    lines and all non-keywords must be quoted with \"s. Comments must be on
    separate lines and start the line with '--'.

    (The bit below is ugly due to a gendoc bug. Sorry.)
    
    Omissions:      
    - LINKTYPE: does not apply to XML
    - SGMLDECL: does not apply to XML
    - OVERRIDE: not fully understood yet
    - SYSTEM:   not fully understood yet
    - DOCUMENT: does not seem too useful
    - BASE:     too complex for now
    - DELEGATE: too complex for now"""

    def __init__(self):
	self.fpi_hash={}
	self.g_entity_hash={}
	self.p_entity_hash={}
	self.doctype_hash={}
	self.not_hash={}
	self.unread_catalogs=[]

    def read_catalog(self,infile):
	"Reads a catalog file from the infile parameter."
	
	while 1:
	    line=infile.readline()
	    if line=="": break

	    # Breaking up the line
	    fields=[]
	    pos=string.index(line," ")
	    fields.append(line[:pos])

	    pos=string.index(line,"\"",pos)+1
	    pos2=string.index(line,"\"",pos)
	    fields.append(line[pos:pos2])

	    pos=string.index(line,"\"",pos2+1)+1
	    pos2=string.index(line,"\"",pos)
	    fields.append(line[pos:pos2])

	    if fields[0]=="PUBLIC":
		self.fpi_hash[fields[1]]=fields[2]
	    elif fields[0]=="ENTITY":
		if fields[1][0]=="%":
		    self.p_entity_hash[fields[1][1:]]=fields[2]
		else:
		    self.g_entity_hash[fields[1]]=fields[2]
	    elif fields[0]=="DOCTYPE":
		self.doctype_hash[fields[1]]=fields[2]
	    elif fields[0]=="NOTATION":
		self.not_hash[fields[1]]=fields[2]
	    elif fields[0]=="CATALOG":
		self.unread_catalogs.append(fields[1])
	    else:
		raise XMLError("Unknown catalog file entry",line[:7])

    def resolve_fpi(self,fpi):
	return self.fpi_hash[fpi]

    def resolveDocumentFPI(self,fpi):
	"Returns the correct system identifier for this document FPI."
	try:
	    return self.fpi_hash[fpi]
	except KeyError,e:
	    return ""

    def resolveDoctypeFPI(self,fpi):
	"Returns the correct system identifier for this document type FPI."
	try:
	    try:
		return self.doctype_hash[fpi]
	    except KeyError,e:
		return self.fpi_hash[fpi]
	except KeyError,e:
	    return ""

    def resolveGE_FPI(self,fpi):
	"Returns the correct system identifier for this general entity FPI."
	try:
	    try:
		return self.g_entity_hash[fpi]
	    except KeyError,e:
		return self.fpi_hash[fpi]
	except KeyError,e:
	    return ""

    def resolvePE_FPI(self,fpi):
	"Returns the correct system identifier for this general entity FPI."
	try:
	    try:
		return self.p_entity_hash[fpi]
	    except KeyError,e:
		return self.fpi_hash[fpi]
	except KeyError,e:
	    return ""

	
# A SAX entity handler that uses the catalog object

class CatalogEntityHandler:
    "Uses a catalog file to resolve entity system identifiers."

    def __init__(catalog):
	self.catalog=catalog

    def resolveEntity(entityName, publicID, systemID):
	"""Given the system identifier (URI) systemID, possibly
	accompanied by an entity name (entityName, with the special
	values '[document]' for the document entity and '[external
	dtd]' for the external DTD subset) and/or public identifier
	(publicID), return the system identifier that the parser
	should use to obtain the entity, or null, to instruct the
	parser to skip the entity (in which case the parser may report
	a validation error). In most cases, this method should return
	the suggested system identifier."""

	if publicID=="":
	    # check if this sysID is mapped to something else
	    return systemID
	
	if entityName=="[document]":
	    sysID=catalog.resolveDocumentFPI(publicID)
	elif entityName=="[external dtd]":
	    sysID=catalog.resolveDoctypeFPI(publicID)
	elif entityName=="":
	    sysID=systemID
	else:
	    sysID=catalog.resolveGE_FPI(publicID)	    

	# check if sysID mapped to something else
	return sysID
	    
    def changeEntity(self,systemID):
	"""Handle a change in the current entity URI. The systemID
	argument specifies the base URI that is now in force."""
	pass
    
# A collection of useful functions

# Utility functions

def unhex(hex_value):
    "Converts a string hex-value to an integer."

    sum=0
    for char in hex_value:
	sum=sum*16
	char=ord(char)
	
	if char<58 and char>=48:
	    sum=sum+(char-48)
	elif char>=97 and char<=102:
	    sum=sum+(char-87)
	elif char>=65 and char<=70:
	    sum=sum+(char-55)
	# else ERROR, but it can't occur here

    return sum

def matches(regexp,str):
    mo=regexp.match(str)
    return mo!=None and len(mo.group(0))==len(str)

# Some useful regexps

reg_ws=re.compile("[\n\t \r]+")
reg_ver=re.compile("\"[-a-zA-Z0-9_.:]+\"|'[-a-zA-Z0-9_.:]+'")
reg_enc_name=re.compile("\"[A-Za-z][-A-Za-z0-9._]*\"|"+\
			"'[A-Za-z][-A-Za-z0-9._]*'")
reg_std_alone=re.compile("\"(yes|no)\"|'(yes|no)'")
reg_comment_content=re.compile("([^-]|-[^-])*")
reg_name=re.compile("[A-Za-z_:][\-A-Za-z_:.0-9]*")
reg_names=re.compile("[A-Za-z_:][\-A-Za-z_:.0-9]*"
		     "([\n\t \r]+[A-Za-z_:][\-A-Za-z_:.0-9]*)*")
reg_nmtoken=re.compile("[\-A-Za-z_:.0-9]+")
reg_nmtokens=re.compile("[\-A-Za-z_:.0-9]+([\n\t \r]+[\-A-Za-z_:.0-9]+)*")
reg_sysid_quote=re.compile("[^\"]*")
reg_sysid_apo=re.compile("[^']*")
reg_pubid_quote=re.compile("[- \n\t\ra-zA-Z0-9'()+,./:=?;!*#@$_%]*")
reg_pubid_apo=re.compile("[- \n\t\ra-zA-Z0-9()+,./:=?;!*#@$_%]*")
reg_start_tag=re.compile("<[A-Za-z_:]")
reg_quoted_attr=re.compile("[^<\"]*")
reg_apo_attr=re.compile("[^<']*")
reg_c_data=re.compile("[<&]")
reg_pe_ref=re.compile("%[A-Za-z_:][\-A-Za-z_:.0-9]*;")

reg_ent_val_quote=re.compile("[^\"]+")
reg_ent_val_apo=re.compile("[^\']+")

reg_attr_type=re.compile(r"CDATA|IDREFS|IDREF|ID|ENTITY|ENTITIES|NMTOKENS|"
			 "NMTOKEN")
reg_attr_def=re.compile(r"#REQUIRED|#IMPLIED")

reg_digits=re.compile("[0-9]+")
reg_hex_digits=re.compile("[0-9a-fA-F]+")

reg_res_pi=re.compile("xml",re.I)

reg_int_dtd=re.compile("\"|'|<\\?|<!--|\\]>|<!\\[")

# RFC 1766 language codes

reg_lang_code=re.compile("([a-zA-Z][a-zA-Z]|[iIxX]-([a-zA-Z])+)(-[a-zA-Z])*")

# Some useful variables

predef_ents={"lt":"&#60;","gt":"&#62;","amp":"&#38;","apos":"&#39;",
             "quot":'&#34;'}

# Translation tables

ws_trans=string.maketrans("\r\t\n","   ")  # Whitespace normalization
id_trans=string.maketrans("","")           # Identity transform 



"""
SAX driver for the Pyexpat C module.
"""

# Note: this driver is 100% experimental and has never been tested at all.
# The following things are missing because I couldn't figure out how to do
# them:
#  - location handling
#
# Event handling can be speeded up by bypassing the driver for some events.
# This will be implemented later when I can test this driver.
#
# This driver has been much improved by Geir Ove Grønmo, who writes:
#
#   "This version of the driver should work, but there are things that
#    need fixing. I have to create a new parser object every time parse_file
#    is called! This is necessary because pyexpat is not reset after parsing."

version="0.10"

from xml.sax import saxlib,saxutils

import urllib
from xml.parsers import pyexpat

# --- SAX_expat

class SAX_expat(saxlib.Parser,saxlib.Locator):
    "SAX driver for the Pyexpat C module."

    def __init__(self):
        saxlib.Parser.__init__(self)
        self.parser=pyexpat.ParserCreate()
        self.parser.StartElementHandler = self.startElement
        self.parser.EndElementHandler = self.endElement
        self.parser.CharacterDataHandler = self.characters
        self.parser.ProcessingInstructionHandler = self.processingInstruction

    def startElement(self,name,attrs):
        at = {}
	for i in range(0, len(attrs), 2):
	    at[attrs[i]] = attrs[i+1]
	    
        self.doc_handler.startElement(name,saxutils.AttributeMap(at))

    def endElement(self,name):
        self.doc_handler.endElement(name)

    def characters(self,data):
        self.doc_handler.characters(data,0,len(data))

    def processingInstruction(self,target,data):
        self.doc_handler.processingInstruction(target,data)

    def parse(self,sysID):
	self.sysID=sysID
        self.parseFile(urllib.urlopen(sysID))
        
    def parseFile(self,fileobj):
        self.reset()
	self.doc_handler.startDocument()

#        while 1:
#            buf=fileobj.read(16384)
#
#            if buf:
#	        if not self.parser.Parse(buf):
#                    self.__report_error()
#                    return
#	    else:
#		break

        if not self.parser.Parse(fileobj.read(),1):
            self.__report_error()
                            
        self.doc_handler.endDocument()

    # --- Locator methods. Only usable after errors.

    def getLineNumber(self):
        return self.parser.ErrorLineNumber

    def getColumnNumber(self):
        return self.parser.ErrorColumnNumber    

    # --- Internal

    def __report_error(self):
        msg=pyexpat.ErrorString(self.parser.ErrorCode)
        self.err_handler.fatalError(saxlib.SAXParseException(msg,None,self))

    # --- EXPERIMENTAL PYTHON SAX EXTENSIONS
        
    def get_parser_name(self):
        return "pyexpat"

    def get_parser_version(self):
        return "Unknown"

    def get_driver_version(self):
        return version
    
    def is_validating(self):
        return 0

    def is_dtd_reading(self):
        return 0

    def reset(self):
        self.parser=pyexpat.ParserCreate()
        self.parser.StartElementHandler = self.startElement
        self.parser.EndElementHandler = self.endElement
        self.parser.CharacterDataHandler = self.characters
        self.parser.ProcessingInstructionHandler = self.processingInstruction
    
    def feed(self,data):
        if not self.parser.Parse(data):
            self.__report_error()

    def close(self):
        if not self.parser.Parse("",1):
            self.__report_error()
    
# ---
        
def create_parser():
    return SAX_expat()



"""
Common code for the sgmllib, htmllib and xmllib parser drivers.
"""

from xml.sax import saxlib,saxutils

import urllib

# --- LibParser

class LibParser(saxlib.Parser,saxlib.Locator):
    "Common code for the sgmllib, htmllib and xmllib parser drivers."

    def __init__(self):
	saxlib.Parser.__init__(self)
	
    def parse(self,sysID):
	"Parses the referenced document."
	self.sysID=sysID
        self.parseFile(urllib.urlopen(sysID))
        
    def parseFile(self,fileobj):
        "Parses the given file."
        if self._can_locate():
            self.doc_handler.setDocumentLocator(self)
        self.reset()
	while 1:
	    buf=fileobj.read(16384)
	    if buf=="": break

	    try:
		self.feed(buf)
	    except RuntimeError,e:
		self.err_handler.fatalError(saxlib.SAXException(str(e),e))
	    
	self.close()

    def unknown_endtag(self,tag):
	"Handles end tags."
	self.doc_handler.endElement(tag)

    def handle_xml(self,encoding,standalone):
        "Remembers whether the document is standalone."
        self.standalone= standalone=="yes"
        
    def handle_data(self,data):
	"Handles PCDATA."
	self.doc_handler.characters(data,0,len(data))

    def handle_cdata(self,data):
	"Handles CDATA marked sections."
	self.doc_handler.characters(data,0,len(data))       

    def syntax_error(self, message):
	"Handles non-fatal errors."
	self.err_handler.error(saxlib.SAXException(message,None))
    
        
# --- SGMLParsers
        
class SGMLParsers(LibParser):
    "Common code for the sgmllib and htmllib parsers."
    
    def handle_pi(self,data):
        "Handles processing instructions."
        # Should we try to parse out the name if there is one?
        self.doc_handler.processingInstruction("",data)

    # handle_starttag is never called!
        
    def unknown_starttag(self,tag,attributes):
	"Handles start tags."
        attrs={}
        for (a,v) in attributes:
            attrs[a]=v
        
	self.doc_handler.startElement(tag,saxutils.AttributeMap(attrs))

    def handle_endtag(self,tag,method):
        "Handles end tags."
        self.doc_handler.endElement(tag)
        
    def unknown_entityref(self,name):
        "Handles entity references by throwing an error."
        self.err_handler.fatalError(saxlib.SAXException("Reference to unknown entity "
                                                 "'%s'" % name,None))

    def unknown_charref(self,no):
        "Handles non-ASCII character references."
        self.err_handler.fatalError(saxlib.SAXException("Reference to unknown character '%d'" % no,None))
        
    def report_unbalanced(self,gi):
        "Reports unbalanced tags."
        self.err_handler.fatalError(saxlib.SAXException("Unbalanced end tag for '%s'" % gi,None))
        
    def _can_locate(self):
        "Internal: returns true if location info is available."
        return 0



"""
SAX driver for the sgmlop parser.
"""

version="0.10"

import sgmlop
from xml.sax import saxlib,saxutils
import urllib

class DHWrapper:

    def __init__(self,real_dh):
        self.real_dh=real_dh

    def __getattr__(self,attr):
        return getattr(self.real_dh,attr)

    def startElement(self,name,attrs):
        self.real_dh.startElement(name,saxutils.AttributeMap(attrs))
        
class Parser(saxlib.Parser):

    def __init__(self):
        saxlib.Parser.__init__(self)
        self.parser = sgmlop.XMLParser()
    
    def setDocumentHandler(self, dh):
	self.parser.register(DHWrapper(dh), 1)
        self.doc_handler=dh

    def parse(self, url):
        self.parseFile(urllib.urlopen(url))
        
    def parseFile(self, file):
	parser = self.parser

	while 1:
	    data = file.read(16384)
	    if not data:
		break
	    parser.feed(data)

	self.close()

    # --- EXPERIMENTAL PYTHON SAX EXTENSIONS

    def get_parser_name(self):
        return "sgmlop"

    def get_parser_version(self):
        return "Unknown"

    def get_driver_version(self):
        return version
    
    def is_validating(self):
        return 0

    def is_dtd_reading(self):
        return 0

    def reset(self):
        self.parser=sgmlop.XMLParser()
    
    def feed(self,data):
        self.parser.feed(data)

    def close(self):
        self.parser.close()
        self.doc_handler.endDocument()
        
# ----

def create_parser():
    return Parser()




"""This is the parts of xmlproc that are specific to validation. They
are an application class that receive data from the parser and a
subclass of the parser object that sets this up."""

# Todo:
#
# - make parser report correct position inside internal entities
# - test position reporting and make it work consistently
# - handle XML decl in external entities
# - clean up entity reference handling in attribute values (parse_string)
# - THINK THIS THROUGH AND MAKE A CONSISTENT POLICY FOR INTERNAL
#   AND EXTERNAL ENTITIES OF DIFFERENT TYPES...
#
# - run through the entire test case collection
#
# - implement notation attributes
# - see if namespaces need to be handled in any way
# - speed-up
#
# - BUGFIX:  normalize whitespace in FPIs
# - BUGFIX:  Handle ../relative sysids correctly
#
# - validate the document type itself:
#   - only allow elem types once in mixed-content models
#   - attr default val: must be legal
#   - standalone doc decl VC
#   - external entity notation: must match declared notation
# - report interoperability violations
# - when fatal error has occurred, don't send more data to app
#
# - WFCs:
#   - Well-Formedness Constraint: Entity Declared
#   - no '<' in attr vals
#
# - move more stuff up into AbstractXML (DTDParser must use app attribute)
# - move entityparser down into xmlutils
#
# - parameter entities in DTD
# - make lexical analyzer separate object?
#
# saxlib:
# - metadata: sax version, saxlib version, parser, parser version(?)
# - list available parsers
# - clean up saxdemo command line interface
# - drivers for htmllib, sgmllib and PPs ESIS parser as well
# - write more extensions
# - simpler interface to parser factory
# - improve XMLTok driver by compiling and testing on Jenufa
# - make a regression test suite for the SAX system and compare the drivers
#
# customization:
# - buffer size
# - catalog file
# - warnings

import urlparse

from xmlproc import *
from xmldtd import *
from xmlapp import *

# ==============================
# The validator class
# ==============================

class XMLValidator:
    """XML parser that validates a document and does some of what is required
    of a validating parser, like adding fixed and default attribute values
    etc."""
    
    def __init__(self):
	self.parser=XMLProcessor()
        self.dtd=CompleteDTD(ErrorHandler(self.parser))
	self.val=ValidatingApp(self.dtd)
	self.parser.set_application(self.val)
        self.parser.dtd=self.dtd
	self.parser.ent=self.dtd

    def parse_resource(self,sysid):
        self.parser.parse_resource(sysid)
        
    def set_application(self,app):
	self.val.set_real_app(app)
	app.set_locator(self.parser)
	
    def set_error_handler(self,err):
	self.parser.set_error_handler(err)
	self.dtd.set_error_handler(err)
	self.val.set_error_handler(err)

    def set_dtd_listener(self,dtd_listener):
	self.parser.set_dtd_listener(dtd_listener)
	
    def get_dtd(self):
        return self.dtd

    def get_current_sysid(self):
	return self.parser.get_current_sysid()

    def get_offset(self):
	return self.parser.get_offset()
	
    def get_line(self):
	return self.parser.get_line()

    def get_column(self):
	return self.parser.get_column()    

# Move this behaviour up into XMLProcessor, hide internal methods
	    
# ==============================
# Application object that checks the document
# ==============================

class ValidatingApp(Application):
    "The object that uses the DTD to actually validate XML documents."

    def __init__(self,dtd):
	self.dtd=dtd
	self.cur_elem=None
	self.cur_state=0
	self.stack=[]
	self.err=ErrorHandler(None)
	self.ids={}
	self.idrefs=[]
	self.realapp=Application()
	
    def set_real_app(self,app):
	self.realapp=app

    def set_locator(self,locator):
	Application.set_locator(self,locator)
	self.realapp.set_locator(locator)

    def set_error_handler(self,err):
	self.err=err

    def handle_start_tag(self,name,attrs):
	decl_root=self.dtd.get_root_elem()
	
	if self.cur_elem!=None:
            if self.cur_state!=-1:
                next=self.cur_elem.next_state(self.cur_state,name)
                if next==0:
                    self.err.error("Element '%s' not allowed here" % name)
                else:
                    self.cur_state=next

	    self.stack.append(self.cur_elem,self.cur_state)
	elif decl_root!=None and name!=decl_root:
	    self.err.error("Document root element '%s' does not match "
			   "declared root element" % name)

	try:
	    self.cur_elem=self.dtd.get_elem(name)
	    try:
		self.cur_state=self.cur_elem.get_start_state()
	    except KeyError,e:
		self.err.error("Bad content model: '%s'. (Internal error.)" %\
			       e)
		self.cur_state=1

	    self.validate_attributes(self.dtd.get_elem(name),attrs)
		
	except KeyError,e:
	    self.err.error("Element '%s' not declared" % name)
	    self.cur_state=-1

	self.realapp.handle_start_tag(name,attrs)
	
    def handle_end_tag(self,name):
	"Notifies the application of end tags (and empty element tags)."
	if self.cur_elem!=None and \
	   not self.cur_elem.final_state(self.cur_state):
	    self.err.error("Element '%s' ended, but not finished" % name)
	
	if self.stack!=[]:
	    (self.cur_elem,self.cur_state)=self.stack[-1]
	    del self.stack[-1]

	self.realapp.handle_end_tag(name)
    
    def handle_data(self,data,start,end):
	"Notifies the application of character data."
	if self.cur_elem!=None:
	    next=self.cur_elem.next_state(self.cur_state,"#PCDATA")

	    if next==0:
		mo=reg_ws.match(data[start:end])
		if mo!=None and len(mo.group(0))==len(data[start:end]):
		    self.realapp.handle_ignorable_data(data,start,end)
		    return
		else:
		    self.err.error("Character data not allowed here")
	    else:
		self.cur_state=next

	self.realapp.handle_data(data,start,end)

    def validate_attributes(self,element,attrs):
	"""Validates the attributes against the element declaration and adds
	fixed and default attributes."""

	# Check the values of the present attributes
	for attr in attrs.keys():
	    try:
		decl=element.get_attr(attr)
	    except KeyError,e:
		self.err.error("Unknown attribute '%s'" % attr)
                return
        
            if decl.type!="CDATA":
                attrs[attr]=string.join(string.split(attrs[attr]))
            
            decl.validate(attrs[attr],self.err)
                
	    if decl.type=="ID":
		if self.ids.has_key(attrs[attr]):
		    self.err.error("ID '%s' appears more than once in "
				   "document" % attrs[attr])
		self.ids[attrs[attr]]=""
	    elif decl.type=="IDREF":
		self.idrefs.append((self.locator.get_line(),
				    self.locator.get_column(),
				    attrs[attr]))
	    elif decl.type=="IDREFS":
		for idref in string.split(attrs[attr]):
		    self.idrefs.append((self.locator.get_line(),
					self.locator.get_column(),
					idref))
	    elif decl.type=="ENTITY":
                try:
                    ent=self.dtd.resolve_ge(attrs[attr])
		    if ent.notation=="":
			self.err.error("Only unparsed entities allowed here")
		    else:
                        try:
                            self.dtd.get_notation(ent.notation)
                        except KeyError,e:
                            self.err.error("Notation '%s' not declared" %\
                                           ent.notation)
                except KeyError,e:
                    self.err.fatal("Unknown entity '%s'" % name)

	    elif decl.type=="ENTITIES":
		for ent_ref in string.split(attrs[attr]):
                    try:
                        ent=self.dtd.resolve_ge(attrs[ent_ref])

			if ent.notation=="":
			    self.err.error("Only unparsed entities allowed "
					   "here")
			else:
                            try:
                                self.dtd.get_notation(ent.notation)
                            except KeyError,e:
                                self.err.error("Notation '%s' not declared" %\
                                               ent.notation)
                    except KeyError,e:
                        self.err.fatal("Unknown entity '%s'" % name)

        # Check for missing required attributes
	for attr in element.get_attr_list():
	    decl=element.get_attr(attr)
	    if decl.decl=="#REQUIRED" and not attrs.has_key(attr):
		self.err.error("Required attribute %s not present" % attr)

    def doc_end(self):
	for (line,col,id) in self.idrefs:
	    if not self.ids.has_key(id):
		self.err.error("IDREF referred to non-existent ID '%s'" % id)

	self.realapp.doc_end()

    def handle_doctype(self,rootname,pub_id,sys_id):
	self.dtd.root_elem=rootname

	prev_loc=self.err.get_locator()
	p=DTDParser()
	p.set_error_handler(self.err)
	p.set_dtd_consumer(self.dtd)

        # === UGLY: SHOULD REALLY ACCESS THE PARSER IN SOME OTHER WAY HERE!
        if self.locator.dtd_listener!=None:
            self.dtd.set_dtd_listener(self.locator.dtd_listener)
	self.err.set_locator(p)

	try:
	    sys_id=urlparse.urljoin(p.get_current_sysid(),sys_id)
	    p.parse_resource(sys_id)
	finally:
	    self.err.set_locator(prev_loc)

	self.realapp.handle_doctype(rootname,pub_id,sys_id)
	    
    # --- These methods added only to make this hanger-on application
    #     invisible to external users.
		
    def doc_start(self):
	self.realapp.doc_start()
	
    def handle_comment(self,data):
	self.realapp.handle_comment(data)

    def handle_ignorable_data(self,data,start,end):
	self.realapp.handle_ignorable_data(data,start,end)

    def handle_pi(self,target,data):
	self.realapp.handle_pi(target,data)

    def set_entity_info(self,xmlver,enc,sddecl):
	self.realapp.set_entity_info(xmlver,enc,sddecl)



"""
SAX driver for xmllib.py
"""

version="0.90"

from xml.sax import saxutils
from xml.sax.drivers import pylibs

import xmllib

# --- SAX_XLParser

class SAX_XLParser(pylibs.LibParser,xmllib.XMLParser):
    "SAX driver for xmllib.py."

    def __init__(self):
	xmllib.XMLParser.__init__(self)
        pylibs.LibParser.__init__(self)
        self.standalone=0
        self.reset()

    def unknown_starttag(self,tag,attributes):
	self.doc_handler.startElement(tag,saxutils.AttributeMap(attributes))
        
    def handle_endtag(self,tag,method):
        self.doc_handler.endElement(tag)

    def handle_proc(self,name,data):
        self.doc_handler.processingInstruction(name,data[1:])

    def getLineNumber(self):
	return self.lineno

    def getSystemId(self):
	return self.sysID

    def _can_locate(self):
        "Internal: returns true if location info is available."
        return 1

    # --- EXPERIMENTAL SAX PYTHON EXTENSIONS

    def get_parser_name(self):
        return "xmllib"

    def get_parser_version(self):
        return xmllib.version

    def get_driver_version(self):
        return version
    
    def is_validating(self):
        return 0

    def is_dtd_reading(self):
        return 0

    def reset(self):
        xmllib.XMLParser.reset(self)
        self.unfed_so_far=1 
    
    def feed(self,data):
        if self.unfed_so_far:
            self.doc_handler.startDocument()
            self.unfed_so_far=0
            
        xmllib.XMLParser.feed(self,data)
    
    def close(self):
        xmllib.XMLParser.close(self)
	self.doc_handler.endDocument()
    
# --- Global functions

def create_parser():
    return SAX_XLParser()




"""
A command-line interface to the xmlproc parser. It continues parsing
after even fatal errors, in order to be find more errors, since this
does not mean feeding data to the application after a fatal error
(which would be in violation of the spec).

Usage:

  [xpcmd.py] urltodoc

urltodoc: URL to the document to parse. (You can use plain file names
          as well.)
"""

# --- INITIALIZATION

import sys
from xml.parsers.xmlproc import xmlproc

# --- ERROR HANDLING

class MyErrorHandler(xmlproc.ErrorHandler):

    def __init__(self,loc):
	self.locator=loc
	self.errors=0
	self.warnings=0

    def get_location(self):
	return "%s:%d:%d" % (self.locator.get_current_sysid(),\
			     self.locator.get_line(),
			     self.locator.get_column())
	
    def warning(self,msg):
	print "WARNING ON %s:" % self.get_location()
	print "  "+msg
	self.warnings=self.warnings+1
    
    def fatal(self,msg):
	print "ERROR ON %s: " % self.get_location()
	print "  "+msg
	self.errors=self.errors+1
	sys.exit(1)

# --- MAIN PROGRAM
	
p=xmlproc.XMLProcessor()
err=MyErrorHandler(p)

p.set_error_handler(err)

print "Parsing document"

p.parse_resource(sys.argv[1])
print "Parse complete, %d error(s) and %d warning(s)" % \
      (err.errors,err.warnings)

# try:
# except:
#     print "Error, parse aborted."



"""A module of experimental extensions to the standard SAX interface."""

import saxlib,imp,sys,string

# --- Internal utility methods

def rec_find_module(module):
    "Improvement over imp.find_module which finds submodules."
    path=""
    for mod in string.split(module,"."):
        if path=="":
            info=(mod,)+imp.find_module(mod)
        else:
            info=(mod,)+imp.find_module(mod,[path])
            
        lastmod=apply(imp.load_module,info)

        try:
            path=lastmod.__path__[0]
        except AttributeError,e:
            pass

    return info

# --- Parser factory

class ParserFactory:
    """A general class to be used by applications for creating parsers on
    foreign systems where it is unknown which parsers exist."""

    def __init__(self,list=None):
        self.parsers=list

    def get_parser_list(self):
        "Returns the list of possible drivers."
        return self.parsers

    def set_parser_list(self,list):
        "Sets the driver list."
        self.parsers=list

    def make_parser(self,drv_name=None):
        """Returns a SAX driver for the first available parser of the parsers
        in the list. Note that the list is one of drivers, so it first tries
        the driver and if that exists imports it to see if the parser also
        exists. If no parsers are available a SAXException is thrown.

        Accepts the driver package name as an optional argument."""

        if drv_name==None:
            list=self.parsers
        else:
            list=[drv_name]
            
        for parser_name in list:
            try:
                info=rec_find_module(parser_name)
                drv_module=apply(imp.load_module,info)
                return drv_module.create_parser()
            except ImportError,e:
                pass

        raise saxlib.SAXException("No parsers found",None)

# --- Experimental extension to Parser interface

class ExtendedParser(saxlib.Parser):
    "Experimental unofficial SAX level 2 extended parser interface."

    def get_parser_name(self):
        "Returns a single-word parser name."
        raise saxlib.SAXException("Method not supported.",None)

    def get_parser_version(self):
        """Returns the version of the imported parser, which may not be the
        one the driver was implemented for."""
        raise saxlib.SAXException("Method not supported.",None)

    def get_driver_version(self):
        "Returns the version number of the driver."
        raise saxlib.SAXException("Method not supported.",None)        
    
    def is_validating(self):
        "True if the parser is validating, false otherwise."
        raise saxlib.SAXException("Method not supported.",None)

    def is_dtd_reading(self):
        """True if the parser is non-validating, but conforms to the spec by
        reading the DTD."""
        raise saxlib.SAXException("Method not supported.",None)

    def reset(self):
        "Makes the parser start parsing afresh."
        raise saxlib.SAXException("Method not supported.",None)
    
    def feed(self,data):
        "Feeds data to the parser."
        raise saxlib.SAXException("Method not supported.",None)

    def close(self):
        "Called after the last call to feed, when there are no more data."
        raise saxlib.SAXException("Method not supported.",None)
        
# --- Experimental document handler which does not slice strings

class NosliceDocumentHandler(saxlib.DocumentHandler):
    """A document handler that does not force the client application to
    slice character data strings."""

    def __init__(self):
        saxlib.DocumentHandler.__init__()
        self.characters=self.safe_handler

    def safe_handler(self,data,start,length):
        """A characters event handler that always works, but doesn't always
        slice strings."""
        if start==0 and length==len(data):
            self.handle_data(data)
        else:
            self.handle_data(data[start:start+length])

    def slice_handler(self,data,start,length):
        "A character event handler that always slices strings."
        self.handle_data(data[start:start+length])

    def noslice_handler(self,data,start,length):
        "A character event handler that never slices strings."
        self.handle_data(data)
        
    def handle_data(self,data):
        "This is the character data event method to override."
        pass

# --- Creating parser factories

XMLParserFactory=ParserFactory(["xml.sax.drivers.drv_pyexpat",
                                "xml.sax.drivers.drv_xmltok",
                                "xml.sax.drivers.drv_xmlproc",
                                "xml.sax.drivers.drv_xmltoolkit",
                                "xml.sax.drivers.drv_xmllib",
                                "xml.sax.drivers.drv_xmldc",
                                "xml.sax.drivers.drv_sgmlop"])

XMLValParserFactory=ParserFactory(["xml.sax.drivers.drv_xmlproc_val"])

HTMLParserFactory=ParserFactory(["xml.sax.drivers.drv_htmllib",
                                 "xml.sax.drivers.drv_sgmlop",
                                 "xml.sax.drivers.drv_sgmllib"])

SGMLParserFactory=ParserFactory(["xml.sax.drivers.drv_sgmlop",
                                 "xml.sax.drivers.drv_sgmllib"])

def make_parser(parser=None):
    if parser==None:
        return XMLParserFactory.make_parser()
    else:
        return XMLParserFactory.make_parser(parser)



"""
A SAX driver for xmlproc
"""

version="0.90"

from xml.sax import saxlib,saxutils
from xml.parsers.xmlproc import xmlproc

# --- SAX_XPParser

class SAX_XPParser(saxlib.Parser,xmlproc.Application,xmlproc.DTDConsumer,
                   xmlproc.ErrorHandler):

    def __init__(self):
	saxlib.Parser.__init__(self)
        self.reset()
    
    def parse(self,sysID):
        self.reset()
	self.parser.parse_resource(sysID)

    def parseFile(self,file):
        self.reset()
	self.parser.parseStart()
	self.parser.read_from(file)
	self.parser.flush()
	self.parser.parseEnd()
        
    def __create_parser(self):
	return xmlproc.XMLProcessor()
    
    def doc_start(self):
	self.doc_handler.startDocument()

    def doc_end(self):
	self.doc_handler.endDocument()

    def handle_data(self,data,start,end):
	self.doc_handler.characters(data,start,end-start)

    def handle_ignorable_data(self,data,start,end):
	self.doc_handler.ignorableWhitespace(data,start,end-start)

    def handle_pi(self, target, data):
	self.doc_handler.processingInstruction(target,data)

    def handle_start_tag(self, name, attrs):
	self.doc_handler.startElement(name,saxutils.AttributeMap(attrs))

    def handle_end_tag(self, name):
	self.doc_handler.endElement(name)

    def resolve_entity(self,pubid,sysid):
        self.ent_handler.resolve_entity(pubid,sysid)
        
    # --- error handling

    def warning(self,msg):
	self.err_handler.warning(saxlib.SAXParseException(msg,None,self))

    def error(self,msg):
	self.err_handler.error(saxlib.SAXParseException(msg,None,self))

    def fatal(self,msg):
	self.err_handler.fatalError(saxlib.SAXParseException(msg,None,self))

    # --- location handling

    def getColumnNumber(self):
	return self.parser.get_column()

    def getLineNumber(self):
	return self.parser.get_line()

    def getSystemId(self):
	return self.parser.get_current_sysid()

    # --- DTD parsing

    def new_external_entity(self,ent_name,sys_id,pub_id,ndata):
        if ndata!="":
            self.dtd_handler.unparsedEntityDecl(ent_name,pub_id,sys_id,ndata)

    def new_notation(self,name,sysid,pubid):
        self.dtd_handler.notationDecl(name,pubid,sysid)

    # --- entity events

    def resolve_entity(self,pubid,sysid):
        newsysid=self.ent_handler.resolveEntity(pubid,sysid)
        if newsysid==None:
            return sysid
        else:
            return newsysid

    # --- EXPERIMENTAL PYTHON SAX EXTENSIONS:

    def get_parser_name(self):
        return "xmlproc"

    def get_parser_version(self):
        return xmlproc.version

    def get_driver_version(self):
        return version
    
    def is_validating(self):
        return 0

    def is_dtd_reading(self):
        return 1

    def reset(self):
	self.parser=self.__create_parser()
	self.parser.set_application(self)
        self.parser.set_dtd_listener(self)
	self.parser.set_error_handler(self)
	self.doc_handler.setDocumentLocator(self)
        self.unfed_so_far=1 
    
    def feed(self,data):
        if self.unfed_so_far:
            self.parser.parseStart()
            self.unfed_so_far=0
            
        self.parser.feed(data)

    def close(self):
        self.parser.flush()
        self.parser.parseEnd()
	
# --- Global functions

def create_parser():
    return SAX_XPParser()
    



"""writer: writer/lineariser classes for dumping DOM tree to file.

"""

from xml.dom.core import *
from xml.dom.walker import Walker
import string, re, sys

def quote(s):
	s = re.sub('&', '&amp;', s)
	s = re.sub('<', '&lt;', s)
	s = re.sub('>', '&gt;', s)
	return s
	

class OutputStream:
	def __init__(self, file):
		self.file = file
		self.new_line = 1

	def write(self, s):
		#print 'write', `s`
		self.file.write(re.sub('\n+', '\n', s))
		if s and s[-1] == '\n':
			self.new_line = 1
		else:
			self.new_line = 0

	def newLine(self):
		if not self.new_line:
			self.write('\n')

	def __del__(self):
		self.file.flush()


class XmlWriter(Walker):

	def __init__(self, stream=sys.stdout):
		self.stream = OutputStream(stream)
		self.empties = []
		self.strip = []
		self.xml_style_endtags = 1
		self.newline_before_start = []
		self.newline_after_start = []
		self.newline_before_end = []
		self.newline_after_end = []
		self.map_attr = self.map_tag = lambda x: x

	def write(self, x):
		if type(x) == type(''):
			self.stream.write(x)
		elif type(x) in (type(()), type([])):
			for y in x:
				self.write(y)
		else:
			self.walk(x)


	def startElement(self, element) :
		assert element.NodeType == ELEMENT

		s = '<%s' % self.map_tag(element.tagName)
		
		# XXX use DOM interface here.
		for name, value in element.attributes.items() :
			s = s + ' %s="%s"' % (self.map_attr(name), quote(value))

		if self.xml_style_endtags and not element.getChildren():
			s = s + '/>'
		else:
			s = s + '>'

		if element.tagName in self.newline_before_start:
			self.stream.newLine()
		self.stream.write(s)
		if element.tagName in self.newline_after_start:
			self.stream.newLine()


	def endElement(self, element):
		assert element.NodeType == ELEMENT

		s = ''
		if element.tagName in self.empties :
			pass
		elif len(element.getChildren()) == 0 and self.xml_style_endtags:
			pass
		else:
			s = s + '</%s>' % self.map_tag(element.tagName)

		if element.tagName in self.newline_before_end:
			self.stream.newLine()
		self.stream.write(s)
		if element.tagName in self.newline_after_end:
			self.stream.newLine()


	def doText(self, text_node):
		#if text_node.getParentNode().tagName in self.strip:
		#	data = string.strip(text_node.data)
		#else:
		data = text_node.data
		self.stream.write(quote(data))


class XmlLineariser(XmlWriter):

	def __init__(self):
		import StringIO
		self.buffer = StringIO.StringIO()
		XmlWriter.__init__(self, self.buffer)

	def linearise(self, node):
		self.write(node)
		return self.buffer.getvalue()
	

class HtmlWriter(XmlWriter):
	def __init__(self, stream=sys.stdout):
		XmlWriter.__init__(self, stream)
		self.map_attr = self.map_tag = string.upper
		self.xml_style_endtags = 0

		self.empties = [
			'img', 'br', 'hr', 'include', 'li', 'meta', 'input',
			'IMG', 'BR', 'HR', 'INCLUDE', 'LI', 'META', 'INPUT',
		]
		self.strip = [
			'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 
			'li', 'br', 'p', 'a', 'title', 'font',
			'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 
			'LI', 'BR', 'P', 'A', 'TITLE', 'FONT',
		]

		nl_dict = {
			'head': (1, 1, 1, 1),
			'body': (1, 1, 1, 1),
			'title': (1, 1, 1, 1),
			'meta': (1, 1, 0, 0),
			'ul': (1, 1, 1, 1),
			'li': (1, 0, 0, 0),
			'h1': (1, 0, 0, 1),
			'h2': (1, 0, 0, 1),
			'h3': (1, 0, 0, 1),
			'h4': (1, 0, 0, 1),
			'h5': (1, 0, 0, 1),
			'h6': (1, 0, 0, 1),
			'p': (1, 0, 0, 1),
			'br': (1, 1, 0, 0),
		}
		for k, v in nl_dict.items():
			if v[0]:
				self.newline_before_start.append(k)
				self.newline_before_start.append(string.upper(k))
			if v[1]:
				self.newline_after_start.append(k)
				self.newline_after_start.append(string.upper(k))
			if v[2]:
				self.newline_before_end.append(k)
				self.newline_before_end.append(string.upper(k))
			if v[3]:
				self.newline_after_end.append(k)
				self.newline_after_end.append(string.upper(k))


class HtmlLineariser(HtmlWriter):

	def __init__(self):
		import StringIO
		self.buffer = StringIO.StringIO()
		HtmlWriter.__init__(self, self.buffer)

	def linearise(self, node):
		self.write(node)
		return self.buffer.getvalue()
	

class ASPWriter(XmlWriter):
	def __init__(self, rep_file):
		self.rep_dict = {}
		self.parseRepFile(rep_file)
		

	def parseRepFile(self, rep_file):
		s = ''
		for l in open(rep_file).readlines():
			if l[0] == '<':
				plus_before = 0
				plus_after = 0
				n = string.index(l, '>')
				tag_name = l[1:n]
				rep = string.strip(l[n+1:])
				if rep and rep[0] == '+':
					plus_before = 1
					rep = string.strip(rep[1:])
				if rep and rep[-1] == '+':
					plus_after = 1
					rep = string.strip(rep[:-1])
				if rep:
					self.rep_dict[tag_name] = (plus_before, plus_after, eval(rep))
				else:
					self.rep_dict[tag_name] = (plus_before, plus_after, '')
					

	def linearise_element(self, element) :
		assert element.NodeType == ELEMENT
		s = ''
		
		# Start tag
		plus_before, plus_after, repl = self.rep_dict[element.getTagName()]

		if s and s[-1] != '\n' and plus_before:
			s = s + '\n'
		s = s + repl
		if s and s[-1] != '\n' and plus_after:
			s = s + '\n'

		s1 = ''
		for child in element.getChildren():
			if child.NodeType is ELEMENT:
				s1 = s1 + self.linearise_element(child)
			elif child.NodeType is TEXT:
				#s1 = s1 + quote(child.data)
				s1 = s1 + child.data
			else :
				s1 = s1 + str(child)
		
		s = s + s1

		# End tag.
		plus_before, plus_after, repl = self.rep_dict['/' + element.getTagName()]
		if s and s[-1] != '\n' and plus_before:
			s = s + '\n'
		s = s + repl
		if s and s[-1] != '\n' and plus_after:
			s = s + '\n'

		return s





# A demo SAX application using the xmllib.py driver, a command-line
# interface and the ESIS outputter.

from xml.sax import saxexts, saxlib, saxutils

import sys,StringIO

in_sysID="quotes.xml"

for klass in [saxutils.Canonizer, saxutils.ESISDocHandler]:
    output = StringIO.StringIO()
    dh = klass(output)
    p=saxexts.make_parser()
    p.setDocumentHandler(dh)
    p.parse(in_sysID)
    print output.__dict__
    print output.getvalue()
    



# benchmark

import time
import sgmlop
from xml.parsers import sgmllib

SIZE = 16384
FILE = "hamlet.xml"

bytes = len(open(FILE).read())

def t1():
    fp = open(FILE)
    parser = sgmllib.SlowSGMLParser()
    while 1:
	data = fp.read(SIZE)
	if not data:
	    break
	parser.feed(data)
    parser.close()
    fp.close()

def t2():
    fp = open(FILE)
    parser = sgmllib.FastSGMLParser()
    while 1:
	data = fp.read(SIZE)
	if not data:
	    break
	parser.feed(data)
    parser.close()
    fp.close()

def t3():
    fp = open(FILE)
    parser = sgmlop.SGMLParser()
    while 1:
	data = fp.read(SIZE)
	if not data:
	    break
	parser.feed(data)
    parser.close()
    fp.close()

class Dummy:
    def finish_starttag(self, tag, data):
	pass
    def finish_endtag(self, tag):
	pass
    def handle_entityref(self, data):
	pass
    def handle_data(self, data):
	pass

def t4():
    fp = open(FILE)
    parser = sgmlop.SGMLParser()
    parser.register(Dummy())
    while 1:
	data = fp.read(SIZE)
	if not data:
	    break
	parser.feed(data)
    parser.close()
    fp.close()

t = time.time()
t1(); t1(); t1(); t1(); t1(); 
t1(); t1(); t1(); t1(); t1(); 
t = (time.time() - t) / 10
print "t1:", t
if t: print int(bytes / t), "bytes per second"

t = time.time()
t2(); t2(); t2(); t2(); t2(); 
t2(); t2(); t2(); t2(); t2(); 
t2(); t2(); t2(); t2(); t2(); 
t2(); t2(); t2(); t2(); t2(); 
t = (time.time() - t) / 20
print "t2:", t
if t: print int(bytes / t), "bytes per second"

t = time.time()
t3(); t3(); t3(); t3(); t3(); 
t3(); t3(); t3(); t3(); t3(); 
t3(); t3(); t3(); t3(); t3(); 
t3(); t3(); t3(); t3(); t3(); 
t = (time.time() - t) / 20
print "t3:", t
if t: print int(bytes / t), "bytes per second"

t = time.time()
t4(); t4(); t4(); t4(); t4(); 
t4(); t4(); t4(); t4(); t4(); 
t4(); t4(); t4(); t4(); t4(); 
t4(); t4(); t4(); t4(); t4(); 
t = (time.time() - t) / 20
print "t4:", t
if t: print int(bytes / t), "bytes per second"



'''
core.py: `light' implementation of the Document Objet Model (core) level 1.

Reference: http://www.w3.org/TR/WD-DOM/level-one-core

Deviations froc the specs:

-- there are no classes NodeList, EditableNodeList and NodeEnumerator,
since node lists are simply implemented as lists.

-- there are no Attributes nor AttributeList classes, as attrubutes are
currently implemented by a dictionnary.

-- there is currently no DocumentContext.

So, useful classes in this module are Node (abstract) and its (concrete)
subclasses -- Document, Element, Text, Comment, PI -- all of
which should be instanciated though a DOMFactory instance.

I have added an alias, ``GI'' for ``tagName'', for compatibility
with previous work.

Typical usage (create document with empty <html> element):

from xml.dom.core import *

dom_factory = DOMFactory()
document = dom_factory.createDocument()
html_node = dom_factory.createElement('html', {})
document.insertBefore(html_node, None)
...

'''

import string

# Exceptions.
NoSuchNodeException = 'No such node'
NotMyChildException = 'Not my child'
NotImplemented = 'Not implemented'

# Node types.

DOCUMENT = 0
ELEMENT = 1
ATTRIBUTE = 1
PI = 3
COMMENT = 4
TEXT = 5

class Node:
	'''Base class for grove nodes in DOM model.'''

	GI = None # Alias for tagName

	def __init__(self, **d):
		self._children = []
		self._parent = None

		for key, value in d.items():
			setattr(self, key, value)
		if hasattr(self, 'tagName'):
			self.GI = self.tagName

	def index(self):
		if self._parent:
			return self._parent._children.index(self)
		else:
			return -1

	# get/set
	def getParentNode(self):
		return self._parent
	
	def getChildren(self):
		# XXX: should return a NodeList.
		return self._children

	def hasChildren(self):
		return len(self._children) > 0

	def getFirstChild(self):
		if self.hasChildren():
			return self._children[0]
		else:
			return None

	def getPreviousSibling(self):
		i = self.index()
		if i <= 0:
			return None
		else:
			return self._parent._children[i - 1]

	def getNextSibling(self):
		i = self.index()
		assert self._parent._children[i] == self
		if i == -1 or i == len(self._parent._children) - 1:
			return None
		else:
			return self._parent._children[i + 1]

	def appendChild(self, new_child):
		'XXX: not in interface.'

		self._children.append(new_child)
		new_child._parent = self
		#new_child.document = self.document

	def insertBefore(self, new_child, ref_child):
		if ref_child == None:
			self.appendChild(new_child)
			return
		for i in range(0, len(self._children)):
			if self._children[i] == ref_child:
				self._children[i:i] = [new_child]
				new_child._parent = self
				#new_child.document = self.document
				return
		raise NotMyChildException
			
	def replaceChild(self, old_child, new_child):
		for i in range(0, len(self._children)):
			if self._children[i] == old_child:
				self._children[i] = new_child
				new_child._parent = self
				return old_child
		raise NotMyChildException
			
	def removeChild(self, old_child):
		try:
			i = self._children.index(old_child)
			del self._children[i]
		except ValueError:
			raise NotMyChildException
			

	def __str__(self):
		from xml.dom.writer import XmlLineariser
		w = XmlLineariser()
		return w.linearise(self)
			

class Document(Node):
	NodeType = DOCUMENT
	GI = "#DOCUMENT"

	def __init__(self):
		Node.__init__(self)
		self.elements_dict = None
		self.documentElement = None
	
	# XXX: experimental, not in DOM.
	#def add_element(self, node):
	#	l = self.elements_dict.get(node.tagName)
	#	if l:
	#		l.append(node)
	#	else:
	#		self.elements_dict[node.tagName] = [node]
	
	def getElementsByTagName(self, name):
		if self.elements_dict is not None:
			return self.elements_dict.get(name, [])
		else:
			return self.documentElement.getElementsByTagName(name)

	def getDocumentElement(self):
	    return self.documentElement

	def setDocumentElement(self, node):
	    self.documentElement = node


class Element(Node):
	NodeType = ELEMENT

	def __init__(self, tagName):
		Node.__init__(self, tagName=tagName)
		self.attributes = {}

	def getTagName(self):
	    return self.tagName
		
	def setAttribute(self, attribute):
		# XXX
		raise NotImplemented

	def getElementsByTagName(self, name):
		# XXX: Return list instead of NodeEnumerator.
		l = []
		for child in self._children:
			if child.NodeType == ELEMENT:
				l = l + child.getElementsByTagName(name)
				if child.tagName == name:
					l.append(child)
		return l

	# XXX: not in DOM.
	def set_attr(self, name, value):
		self.attributes[name] = value


class Attribute(Node):
	# XXX: not used!!!
	NodeType = ATTRIBUTE
	GI = "#ATTR"


class Text(Node):
	GI = "#PCDATA"
	NodeType = TEXT

	def __str__(self):
		return self.data


class Comment(Node):
	GI = "#COMMENT"
	NodeType = COMMENT
	data = ''

	def __str__(self):
		return '<-- ?? -->' % self.data


class PI(Node):
	GI = "#PI"
	NodeType = PI
	data = ''

	def __str__(self):
		return '<? %s ?>' % self.data


# Auxiliary types.

class DOMFactory:
	def createDocumentContext(self):
		return DocumentContext()

	def createDocument(self):
		return Document()

	def createElement(self, tagName, attributes):
		e = Element(tagName)
		e.attributes = attributes
		return e

	def createAttribute(self, name, value):
		a = Attribute()
		a.name = name
		a.value = value
		
	def createTextNode(self, data):
		return Text(data=data)
		
	def createComment(self, data):
		return Comment(data=data)
		

# vim:ts=2:ai



"""This file (which is sourced, not imported) checks the version of the
"versioncheck" package. It is also an example of how to format your own
_checkversion.py file"""

import pyversioncheck

_PACKAGE="xml"
_VERSION="0.3"
_URL="http://www.python.org/sigs/xml-sig/files/xmlpackageversion.txt"

try:
	_myverbose=VERBOSE
except NameError:
	_myverbose=1
	
pyversioncheck.versioncheck(_PACKAGE, _URL, _VERSION, verbose=_myverbose)



"""HTML elements.

Clases representing HTML4.0 elements, subclassed from DOM.core.Element.
"""

from pydom import makeElementClass

def makeAll():
	empties = ['BR', 'HR', 'IMG']
	non_empties = [
		'HTML', 
		'HEAD', 'META', 'TITLE',
		'BODY', 
		'H1', 'H2', 'H3', 'H4', 'H5', 'H6',
		'FONT', 'BASEFONT', 'ADDRESS', 'DIV',
		'CENTER', # deprecated
		'A', 'MAP', 'AREA', 'LINK',
		'APPLET', 'OBJECT', 'SCRIPT',
		'PRE', 
		'P',
		'OL', 'UL', 'LI'
	]
	
	for name in empties:
		c = makeElementClass(name)
		c.is_empty = 1
		globals()[name] = c
	
	for name in non_empties:
		c = makeElementClass(name)
		globals()[name] = c
	
makeAll()
del makeAll

if __name__ == '__main__':
	t = HTML(HEAD(), BODY(H1('this is a test'), "don't panic."))
	print t.getChildren()
	from writer import HtmlWriter, XmlWriter, XmlLineariser

	l = XmlWriter()
	l.write(t)




# 
# Top-level program for XML test suite
#

import sys, regrtest, getopt, string
from test import test_support

def main():
    # This main function is essentially the same as the one in the
    # original regrtest.py shipped with Python 1.5.1, slightly
    # modified to handle running outside of regrtest.  

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'vgqx')
    except getopt.error, msg:
        print msg
        print __doc__
        return 2
    verbose = 0
    quiet = 0
    generate = 0
    exclude = 0
    for o, a in opts:
        if o == '-v': verbose = verbose+1
        if o == '-q': quiet = 1; verbose = 0
        if o == '-g': generate = 1
        if o == '-x': exclude = 1
    if generate and verbose:
        print "-g and -v don't go together!"
        return 2
    good = []
    bad = []
    skipped = []
    for i in range(len(args)):
        # Strip trailing ".py" from arguments
        if args[i][-3:] == '.py':
            args[i] = args[i][:-3]
    if exclude:
        nottests[:0] = args
        args = []
    tests = regrtest.findtests(testdir = '.', stdtests = [], nottests = [])
    test_support.verbose = verbose      # Tell tests to be moderately quiet
    for test in tests:
        if not quiet:
            print test
        ok = regrtest.runtest(test, generate, verbose)
        if ok > 0:
            good.append(test)
        elif ok == 0:
            bad.append(test)
        else:
            if not quiet:
                print "test", test,
                print "skipped -- an optional feature could not be imported"
            skipped.append(test)
    if good and not quiet:
        if not bad and not skipped and len(good) > 1:
            print "All",
        print regrtest.count(len(good), "test"), "OK."
    if bad:
        print regrtest.count(len(bad), "test"), "failed:",
        print string.join(bad)
    if skipped and not quiet:
        print regrtest.count(len(skipped), "test"), "skipped:",
        print string.join(skipped)
    return len(bad) > 0
    
if __name__ == '__main__':
    main()



'''
Event-driven, almost-SAXish, grove builder.


'''

from xml.dom.core import *

class Builder:

	def __init__(self, factory=None):
		self.dom_factory = factory or DOMFactory()
		self.document = self.dom_factory.createDocument()
		self.current_element = None


	def push(self, node):
		"Add node to current node and move to new node."

		if self.current_element:
			self.current_element.insertBefore(node, None)
		elif node.NodeType == ELEMENT:
			self.document.setDocumentElement(node)

		if node.NodeType == ELEMENT:
			self.current_element = node

	def pop(self):
		"Move to current node's parent."

		self.current_element = self.current_element.getParentNode()
		

	def startElement(self, name, attrs):
		if hasattr(self, 'start_' + name):
			getattr(self, 'start_' + name)(elm)
		else:
			element = self.dom_factory.createElement(name, attrs)
			self.push(element)
	
	def endElement(self, name):
		assert name == self.current_element.tagName
		self.pop()


	def text(self, s):
		if self.current_element:
			text_node = self.dom_factory.createTextNode(s)
			self.current_element.insertBefore(text_node, None)





"""A library of useful subclasses to the saxlib classes, for the
convenience of application and driver writers."""

import types,saxlib,string,sys,saxexts

# --- Location

class Location:
    """Represents a location in an XML entity. Initialized by being passed
    a locator, from which it reads off the current location, which is then
    stored internally."""

    def __init__(self, locator):
	self.__col=locator.getColumnNumber()
	self.__line=locator.getLineNumber()
	self.__pubid=locator.getPublicId()
	self.__sysid=locator.getSystemId()
    
    def getColumnNumber(self):
	return self.__col

    def getLineNumber(self):
	return self.__line

    def getPublicId(self):
	return self.__pubid

    def getSystemId(self):
	return self.__sysid

# --- ErrorPrinter
    
class ErrorPrinter:
    "A simple class that just prints error messages to standard out."

    def __init__(self,level=0,outfile=sys.stderr):
        self.level=level
        self.outfile=outfile
    
    def error(self, exception):
        if self.level>1:
            return
        
	self.outfile.write("ERROR in %s: %s\n" % (self.__getpos(exception),\
                                                  exception.getMessage()))

    def fatalError(self, exception):
        if self.level>2:
            return
        
	self.outfile.write("FATAL ERROR in %s: %s\n" % \
                           (self.__getpos(exception),exception.getMessage()))

    def warning(self, exception):
        if self.level>0:
            return
        
	self.outfile.write("WARNING in %s: %s\n" % (self.__getpos(exception),\
                                                    exception.getMessage()))

    def __getpos(self, exception):
	return "%s:%s:%s" % (exception.getSystemId(),\
			     exception.getLineNumber(),\
			     exception.getColumnNumber())

# --- ErrorRaiser

class ErrorRaiser:
    "A simple class that just raises the exceptions it is passed."

    def error(self, exception):
        raise exception

    def fatalError(self, exception):
        raise exception

    def warning(self, exception):
        raise exception
    
# --- AttributeMap

class AttributeMap:
    """An implementation of AttributeList that takes an (attr,val) hash
    and uses it to implement the AttributeList interface."""    

    def __init__(self, map):
	self.map=map
    
    def getLength(self):
	return len(self.map.keys())
	
    def getName(self, i):
        try:
            return self.map.keys()[i]
        except IndexError,e:
            return None

    def getType(self, i):
	return "CDATA"

    def getValue(self, i):
        try:
            if type(i)==types.IntType:
                return self.map[self.getName(i)]
            else:
                return self.map[i]
        except KeyError,e:
            return None

    def __len__(self):
	return len(self.map.keys())

    def __getitem__(self, key):
	if type(key)==types.IntType:
            return self.map.keys()[key]
	else:
            return self.map[key]

    def items(self):
        return self.map.items()
        
    def keys(self):
	return self.map.keys()

    def has_key(self,key):
	return self.map.has_key(key)

# --- Event broadcasting object

class EventBroadcaster:    
    """Takes a list of objects and forwards any method calls received
    to all objects in the list. The attribute list holds the list and
    can freely be modified by clients."""

    class Event:
        "Helper objects that represent event methods."

        def __init__(self,list,name):
            self.list=list
            self.name=name
        
        def __call__(self,*rest):
            for obj in self.list:
                apply(getattr(obj,self.name), rest)
    
    def __init__(self,list):
        self.list=list

    def __getattr__(self,name):
        return self.Event(self.list,name)

    def __repr__(self):
        return "<EventBroadcaster instance at %d>" % id(self)

# ESIS document handler

class ESISDocHandler(saxlib.HandlerBase):
    "A SAX document handler that produces naive ESIS output."

    def __init__(self,writer=sys.stdout):
	self.writer=writer
    
    def processingInstruction (self,target, remainder):
	"""Receive an event signalling that a processing instruction
	has been found."""
	self.writer.write("?"+target+" "+remainder+"\n")

    def startElement(self,name,amap):
	"Receive an event signalling the start of an element."
	self.writer.write("("+name+"\n")
	for a_name in amap.keys():
	    self.writer.write("A"+a_name+" "+amap[a_name]+"\n")

    def endElement(self,name):
	"Receive an event signalling the end of an element."
	self.writer.write(")"+name+"\n")

    def characters(self,data,start_ix,length):
	"Receive an event signalling that character data has been found."
	self.writer.write("-"+data[start_ix:start_ix+length]+"\n")

    def endDocument(self):
        try:
            self.writer.close()
        except NameError:
            pass # It's OK, if the method isn't there we probably don't need it
        
# XML canonizer

class Canonizer(saxlib.HandlerBase):
    "A SAX document handler that produces canonized XML output."

    def __init__(self,writer=sys.stdout):
	self.elem_level=0
	self.writer=writer
    
    def processingInstruction (self,target, remainder):
	if not target=="xml":
	    self.writer.write("<?"+target+" "+remainder+"?>")

    def startElement(self,name,amap):
	self.writer.write("<"+name)
	
	a_names=amap.keys()
	a_names.sort()

	for a_name in a_names:
	    self.writer.write(" "+a_name+"=\"")
	    self.write_data(amap[a_name])
	    self.writer.write("\"")
	self.writer.write(">")
	self.elem_level=self.elem_level+1

    def endElement(self,name):
	self.writer.write("</"+name+">")
	self.elem_level=self.elem_level-1

    def ignorableWhitespace(self,data,start_ix,length):
	self.characters(data,start_ix,length)
	
    def characters(self,data,start_ix,length):
	if self.elem_level>0:
            self.write_data(data[start_ix:start_ix+length])
	    
    def write_data(self,data):
	"Writes datachars to writer."
	data=string.replace(data,"&","&amp;")
	data=string.replace(data,"<","&lt;")
	data=string.replace(data,"\"","&quot;")
	data=string.replace(data,">","&gt;")
        data=string.replace(data,chr(9),"&#9;")
        data=string.replace(data,chr(10),"&#10;")
        data=string.replace(data,chr(13),"&#13;")
	self.writer.write(data)
	
    def endDocument(self):
        try:
            self.writer.close()
        except NameError:
            pass # It's OK, if the method isn't there we probably don't need it

# --- mllib

# Unsupported:
# - setnomoretags
# - setliteral
# - translate_references
# - handle_xml
# - handle_doctype
# - handle_charref
# - handle_entityref
# - handle_comment
# - handle_cdata
# - tag_attributes

class mllib:
    """A re-implementation of the htmllib, sgmllib and xmllib interfaces as a
    SAX DocumentHandler."""

    def __init__(self):
        self.reset()

    def reset(self):
        self.parser=saxexts.XMLParserFactory.make_parser()
        self.handler=mllib.Handler(self.parser,self)
        self.handler.reset()

    def feed(self,data):
        self.parser.feed(data)

    def close(self):
        self.parser.close()

    def get_stack(self):
        return self.handler.get_stack()

    # --- Handler methods (to be overridden)

    def handle_starttag(self,name,method,atts):
        method(atts)

    def handle_endtag(self,name,method):
        method()
    
    def handle_data(self,data):
        pass

    def handle_proc(self,target,data):
        pass

    def unknown_starttag(self,name,atts):
        pass
    
    def unknown_endtag(self,name):
        pass

    def syntax_error(self,message):
        pass
    
    # --- The internal handler class
      
    class Handler(saxlib.DocumentHandler,saxlib.ErrorHandler):
        """An internal class to handle SAX events and translate them to mllib
        events."""

        def __init__(self,driver,handler):
            self.driver=driver
            self.driver.setDocumentHandler(self)
            self.driver.setErrorHandler(self)
            self.handler=handler
            self.reset()

        def get_stack(self):
            return self.stack

        def reset(self):
            self.stack=[]

        # --- DocumentHandler methods
            
        def characters(self, ch, start, length):
            self.handler.handle_data(ch[start:start+length])

        def endElement(self, name):
            if hasattr(self.handler,"end_"+name):
                self.handler.handle_endtag(name,
                                          getattr(self.handler,"end_"+name))
            else:
                self.handler.unknown_endtag(name)

            del self.stack[-1]

        def ignorableWhitespace(self, ch, start, length):
            self.handler.handle_data(ch[start:start+length])

        def processingInstruction(self, target, data):
            self.handler.handle_proc(target,data)

        def startElement(self, name, atts):
            self.stack.append(name)
            
            if hasattr(self.handler,"start_"+name):
                self.handler.handle_starttag(name,
                                            getattr(self.handler,
                                                    "start_"+name),
                                             atts)
            else:
                self.handler.unknown_starttag(name,atts)
            
        # --- ErrorHandler methods

        def error(self, exception):
            self.handler.syntax_error(str(exception))

        def fatalError(self, exception):
            raise RuntimeError(str(exception))



"""This is a validating XML parser with a SAX driver. It currently
handles most of the well-formedness checking and validation and the
rest will be added as soon as possible.

Version 0.40 - 16.May.98
   Lars Marius Garshol - larsga@ifi.uio.no
   http://www.stud.ifi.uio.no/~larsga/download/python/xml/xmlproc.html
"""
   
import re,string,sys,urllib,urlparse,types

from xmlutils import *
from xmlapp import *
from xmldtd import *

version="0.40"

# Standard exceptions

class OutOfDataException(Exception):
    """An exception that signals that more data is expected, but the current
    buffer has been exhausted."""
    pass

# ==============================
# The general entity parser
# ==============================

class EntityParser:
    """A generalized parser for XML entities, whether DTD, documents or even
    catalog files."""

    def __init__(self):
	self.ent_stack=[]
	self.open_ents=[]  # Used to test for entity recursion
	self.app=Application()
	self.err=ErrorHandler(self)
	self.ent=EntityHandler(self.err)

	# Block information
	self.data=""
	self.final=0
	self.datasize=0
	self.start_point=-1
	
	# Location tracking
	self.line=1
	self.last_break=0
	self.block_offset=0 # Offset from start of stream to start of cur block
	self.pos=0
	self.current_sysID=None	
	self.last_upd_pos=0

    def set_application(self,app):
	"Sets the object to send data events to."
	self.app=app
	app.set_locator(self)

    def set_error_handler(self,err):
	"Sets the object to send error events to."
	self.err=err
	self.ent.set_error_handler(err)

    def set_entity_handler(self,ent):
	"Sets the object that resolves entity references."
	self.ent=ent
	self.ent.set_error_handler(self.err)
	
    def parse_resource(self,sysID):
	"""Begin parsing an XML entity with the specified public and
	system identifiers (the system identifier, a URI, is required).
	Only used for the document entity, not to handle subentities, which
	open_entity takes care of."""

	self.current_sysID=sysID
	try:
	    infile=urllib.urlopen(sysID)
	except IOError,e:
	    self.err.fatal("Couldn't open resource '%s'" % sysID)
	    return
	
	self.parseStart()
	self.read_from(infile)
	infile.close()
	self.flush()
	self.parseEnd()

    def open_entity(self,sysID):
	"""Starts parsing a new entity, pushing the old onto the stack. This
	method must not be used to start parsing, use parse_resource for
	that."""

	sysID=urlparse.urljoin(self.get_current_sysid(),sysID)
	    
	try:
	    inf=urllib.urlopen(sysID)
	except IOError,e:
	    self.err.fatal("Couldn't open entity '%s'" % sysID)
	    return

	self.ent_stack.append(self.get_current_sysid(),self.data,self.pos,\
			      self.line,self.last_break,self.datasize,\
			      self.last_upd_pos,self.block_offset)
	
	self.current_sysID=sysID
	self.pos=0
	self.line=1
	self.last_break=0
	self.data=""
	
	self.read_from(inf)

	self.flush()
	self.pop_entity()

    def push_entity(self,sysID,contents):
	"""Parse some text and consider it a new entity, making it possible
	to return to the original entity later."""
	self.ent_stack.append(self.get_current_sysid(),self.data,self.pos,\
			      self.line,self.last_break,self.datasize,\
			      self.last_upd_pos,self.block_offset)

	self.data=contents
	self.current_sysID=sysID
	self.pos=0
	self.line=1
	self.last_break=0
	self.datasize=len(contents)
	self.last_upd_pos=0

    def pop_entity(self):
	"Skips out of the current entity and back to the previous one."

	if self.ent_stack==[]:
	    self.err.fatal("Internal error: Entity stack broken")

	(self.current_sysID,self.data,self.pos,self.line,self.last_break,\
	 self.datasize,self.last_upd_pos,self.block_offset)=self.ent_stack[-1]

	del self.ent_stack[-1]
	self.final=0
	
    def read_from(self,fileobj,bufsize=16384):
	"""Reads data from a file-like object until EOF. Does not close it.
	**WARNING**: This method does not call the parseStart/parseEnd methods,
	since it does not know if it may be called several times. Use
	parse_resource if you just want to read a file."""
	while 1:
	    buf=fileobj.read(bufsize)
	    if buf=="": break
	    self.feed(buf)

    def feed(self,new_data):
	"""Accepts more data from the data source. This method must
	set self.datasize and correctly update self.pos and self.data."""
	self.update_pos() # Update line/col count

        # Doing line end translation
        new_data=string.translate(new_data,id_trans,"\015")
        
	if self.start_point!=-1:
	    self.block_offset=self.block_offset+self.datasize
	    self.data=self.data[self.pos:]+new_data
	    self.last_break=self.last_break-self.pos  # Keep track of column
	    self.pos=0
	    self.last_upd_pos=0
	else:
	    self.data=self.data+new_data
	self.datasize=len(self.data)

	self.do_parse()

    def parseStart(self):
	"Called before the parse starts to notify subclasses."
	pass

    def parseEnd(self):
	"Called when there are no more data to notify subclasses."
	pass

    def flush(self):
	"Parses any remnants of data in the last block."
	if not self.pos+1==self.datasize:
	    self.final=1
 	    try:
		self.do_parse()
	    except OutOfDataException,e:
		self.err.fatal("Construct started, but never completed")
		    
    # --- LOW-LEVEL SCANNING METHODS

    def set_start_point(self):
	"""Stores the current position and tells the parser not to forget any
	of the data beyond this point until get_region is called."""
	self.start_point=self.pos

    def get_region(self):
	"""Returns the area from start_point to current position and remove
	start_point."""
	data=self.data[self.start_point:self.pos]
	self.start_point=-1
	return data

    def find_reg(self,regexp):
	"""Moves self.pos to the first character that matches the regexp and
	returns everything from pos and up to (but not including) that
	character."""
	oldpos=self.pos
	mo=regexp.search(self.data,self.pos)
	if mo==None:
	    raise OutOfDataException()
	self.pos=mo.start(0)
	return self.data[oldpos:self.pos]
    
    def scan_to(self,target):
	"Moves self.pos to beyond target and returns skipped text."
	new_pos=string.find(self.data,target,self.pos)
	if new_pos==-1:
	    raise OutOfDataException()
	res=self.data[self.pos:new_pos]
	self.pos=new_pos+len(target)
	return res

    def get_index(self,target):
	"Finds the position where target starts and returns it."
	new_pos=string.find(self.data,target,self.pos)
	if new_pos==-1:
	    raise OutOfDataException()
	return new_pos
    
    def test_str(self,test_str):
	"See if text at current position matches test_str, without moving."
	if self.datasize-self.pos<len(test_str) and not self.final:
	    raise OutOfDataException()
	return self.data[self.pos:self.pos+len(test_str)]==test_str
    
    def now_at(self,test_str):
	"Checks if we are at this string now, and if so skips over it."
	if self.datasize-self.pos<len(test_str) and not self.final:
	    raise OutOfDataException()
	
	if self.data[self.pos:self.pos+len(test_str)]==test_str:
	    self.pos=self.pos+len(test_str)
	    return 1
	else:
	    return 0
    
    def skip_ws(self,necessary=0):
	"Skips over any whitespace at this point."
	ws=reg_ws.match(self.data,self.pos)
	if ws==None:
	    if necessary:
		self.err.fatal("Whitespace expected here")
	    return
	self.pos=ws.end(0)
	ws=ws.group(0)

    def test_reg(self,regexp):
	"Checks if we match the regexp."
	if self.pos>self.datasize-5 and not self.final:
	    raise OutOfDataException()
	
	return regexp.match(self.data,self.pos)!=None
	    
    def get_match(self,regexp):
	"Returns the result of matching the regexp and advances self.pos."
	if self.pos>self.datasize-5 and not self.final:
	    raise OutOfDataException()
	
	ent=regexp.match(self.data,self.pos)
	if ent==None:
	    self.err.fatal("Didn't match "+regexp.pattern)
	    return ""

	if ent.end(0)==self.datasize:
	    raise OutOfDataException()

	self.pos=ent.end(0)
	return ent.group(0)

    def update_pos(self):
	"Updates (line,col)-pos by checking processed blocks."
	breaks=string.count(self.data[self.last_upd_pos:self.pos],"\n")
	self.last_upd_pos=self.pos

	if breaks>0:
	    self.line=self.line+breaks
	    self.last_break=string.rfind(self.data,"\n",0,self.pos)

    def get_wrapped_match(self,wraps):
	"Returns a contained match. Useful for regexps inside quotes."
	found=0
	for (wrap,regexp) in wraps:
	    if self.test_str(wrap):
		found=1
		self.pos=self.pos+len(wrap)
		break

	if not found:
	    msg="One of "
	    for (wrap,regexp) in wraps[:-1]:
		msg="%s'%s', " % (msg,wrap)
	    self.err.fatal("%s or '%s' expected" % (msg[:-2],wraps[-1][0]))

	data=self.get_match(regexp)
	if not self.now_at(wrap):
	    self.err.fatal("'%s' expected" % (wrap))

	return data
	    
    #--- USEFUL METHODS

    def get_current_sysid(self):
	"Returns the sysid of the file we are reading now."
	return self.current_sysID

    def set_sysid(self,sysID):
	"Sets the current system identifier. Does not store the old one."
	self.current_sysID=sysID

    def get_offset(self):
	"Returns the current offset from the start of the stream."
	return self.block_offset+self.pos
	
    def get_line(self):
	"Returns the current line number."
	self.update_pos()
	return self.line

    def get_column(self):
	"Returns the current column position."
	self.update_pos()
	return self.pos-self.last_break

# ==============================
# Code-sharing class for XMLProcessor and DTDParser
# ==============================

class AbstractXML(EntityParser):
    "An abstract superclass with shared code for XMLProcessor and DTDParser."

    def parse_string(self,where):
	"""Parses a string (attribute value, entity value) and resolves all
	entity references in it. where=0 means entity declaration, where=1
	means attribute string and where=2 means we are parsing the literal
        entity value of a reference to an internal entity in an attribute
        value."""

	val=""
        if where!=2:
            if self.now_at('"'):
                delim='"'
            elif self.now_at("'"):
                delim="'"
            else:
                self.err.fatal("Expected \" or '")
                self.scan_to(">")
                return

	if where==0:
	    reg_stop=re.compile("%|&|"+delim)
	elif where==1:
	    reg_stop=re.compile("<|&|"+delim) # Parameter entities not recogn.
        elif where==2:
            reg_stop=re.compile("<|&")
	        
        while 1:
	    piece=self.find_reg(reg_stop)
            
            if where==1 or where==2:
                val=val+string.translate(piece,ws_trans)
            else:
                val=val+piece

	    if where!=2 and self.now_at(delim):
                break

	    if self.now_at("&#"):
		if self.now_at("x"):
		    digs=unhex(self.get_match(reg_hex_digits))
		else:
		    digs=string.atoi(self.get_match(reg_digits))
		    
		if not (digs==9 or digs==10 or digs==13 or \
			(digs>=32 and digs<=255)):
		    if digs>255:
			self.err.fatal("Unsupported character in character "
				       "reference")
		    else:
			self.err.fatal("Illegal character in character "
				       "reference")
		else:
		    val=val+chr(digs)
		
	    elif self.now_at("&"):	
		name=self.get_match(reg_name)

		if where==0:
		    val=val+"&"+name+";"
		else:
                    try:
                        ent=self.ent.resolve_ge(name)
			if ent.is_internal():
                            # Doing all this here sucks a bit, but...
                            self.push_entity(self.get_current_sysid(),\
                                             ent.value)

                            self.final=1 # Only one block
                            
                            try:
                                val=val+self.parse_string(2)
                            except OutOfDataException,e:
                                print "OODE"
                            
                            if not self.pos==self.datasize:
                                self.err.fatal("Construct started, but never "
                                               "completed" %\
                                               (self.pos+1,self.datasize))

                            self.pop_entity()
			else:
			    self.err.fatal("External entity references not "
					   "allowed here")
                    except KeyError,e:
                        self.err.fatal("Unknown entity '%s'" % name)	       
                        
	    elif self.now_at("%"):
		# We never get here if we are in an attribute value, since
		# % is not included in the reg_stop regexp.
		name=self.get_match(reg_name)

		if self.internal:
		    self.err.fatal("Parameter entity references not "
				       "allowed in internal subset")
		    val=val+"%"+name+";"
		else:
                    try:
                        ent=self.dtd.resolve_pe(name)
			if ent.is_internal():
			    val=val+ent.value
			else:
			    self.err.fatal("External entity references not "
                                           "allowed here")
                    except KeyError,e:
                        self.err.fatal("Unknown parameter entity '%s'" % name)
            elif self.now_at("<") and (where==1 or where==2):
                self.err.fatal("'<' not allowed in attribute values")
	    else:
		self.err.fatal("Entity reference expected. (Internal error.)")
		
	    if not self.now_at(";"):
		self.err.fatal("';' expected")
		self.scan_to(delim)

            if where==2 and self.pos==self.datasize:
                break
                            
	return val
    
    def parse_comment(self):
	"Parses the comment from after '<!--' and beyond '-->'."
	self.app.handle_comment(self.get_match(reg_comment_content))
	if not self.now_at("-->"):
	    self.err.fatal("Comment incorrectly terminated")

    def parse_pi(self):
	"""Parses a processing instruction from after the '<?' and beyond
	the '?>'."""
	trgt=self.get_match(reg_name)

	if trgt=="xml":
	    self.parse_xml_decl()
	    if not self.now_at("?>"):
		self.err.fatal("'?>' expected")
	    self.seen_xmldecl=1
	else:
	    self.skip_ws()
	    rem=self.scan_to("?>") # OutOfDataException if not found

	    if reg_res_pi.match(trgt)!=None:
		if trgt=="xml:namespace":
		    self.err.warning("XML name spaces not yet supported")
		else:
		    self.err.fatal("Processing instruction target names "
				   "beginning with 'xml' are reserved")
	    self.app.handle_pi(trgt,rem)   
	    
# ==============================
# A full well-formedness parser
# ==============================

class XMLProcessor(AbstractXML):
    "A parser that performs a complete well-formedness check."

    def __init__(self):
	EntityParser.__init__(self)

	# State vars
	self.stack=[]
	self.seen_root=0
	self.seen_doctype=0
	self.seen_xmldecl=0

	# Various handlers
	self.dtd=WFCDTD(self.err)
	self.ent=self.dtd
        self.dtd_listener=None
    
    def do_parse(self):
	"Does the actual parsing."
	try:
	    while self.pos+1<self.datasize:
		prepos=self.pos

		if self.test_reg(reg_start_tag):
		    self.parse_start_tag()
		elif self.now_at("</"):
		    self.parse_end_tag()
		elif not (self.test_str("<") or self.test_str("&")):
		    self.parse_data()
		elif self.now_at("&#"):
		    self.parse_charref()
		elif self.now_at("&"):
		    self.parse_ent_ref()
		elif self.now_at("<!--"):
		    self.parse_comment()
		elif self.now_at("<?"):
		    self.parse_pi()
		elif self.now_at("<![CDATA["):
		    self.parse_cdata()
		elif self.now_at("<!DOCTYPE"):
		    self.parse_doctype()
		else:
		    self.err.fatal("Illegal construct")
		    self.pos=self.pos+1 # Avoid endless loop

	except OutOfDataException,e:
	    if self.final:
		raise sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
	        #raise e
	    else:
		self.pos=prepos  # Didn't complete the construct

    def set_dtd_listener(self,listener):
        "Registers an object that listens for DTD parse events."
        self.dtd_listener=listener
                
    def parseStart(self):
	"Must be called before parsing starts. (Notifies application.)"
	self.app.doc_start()

    def parseEnd(self):
	"""Must be called when parsing is finished. (Does some checks and "
	"notifies the application.)"""	    
	if self.stack!=[] and self.ent_stack==[]:
	    self.err.fatal("Premature document end, element '%s' not "
			   "closed" % (self.stack[-1]))
	elif not self.seen_root:
	    self.err.fatal("Premature document end, no root element")

	self.app.doc_end()
	    
    def parse_start_tag(self):
	"Parses the start tag."
	self.pos=self.pos+1 # Skips the '<'
	name=self.get_match(reg_name)
	self.skip_ws()

	attrs={}
	while not (self.test_str("/>") or self.test_str(">")):
	    a_name=self.get_match(reg_name)
	    self.skip_ws()
	    if not self.now_at("="):
		self.err.fatal("'=' expected")
	    self.skip_ws()

	    a_val=self.parse_string(1) # Substitutes entities etc

	    if attrs.has_key(a_name):
		self.err.fatal("Attribute '%s' occurs twice" % a_name)
		
	    attrs[a_name]=a_val		    
	    self.skip_ws()
            
        # --- Insert default attributes
        try:
            element=self.dtd.get_elem(name)
            for attr in element.get_attr_list():
                decl=element.get_attr(attr)

                if decl.decl=="#FIXED":
                    try:
                        if attrs[attr]!=decl.default:
                            self.err.error("Actual value of attribute '%s' "
                                           "does not match fixed value" % attr)
                    except KeyError,e:
                        attrs[attr]=decl.default
                elif decl.decl=="#DEFAULT" and not attrs.has_key(attr):
                    attrs[attr]=decl.default
        except KeyError,e:
            pass
        
	# --- Take care of the tag
	    
	if self.stack==[] and self.seen_root:
	    self.err.fatal("Elements not allowed outside root element")
	    
	self.seen_root=1
	self.stack.append(name)
	    
	if self.now_at(">"):
	    self.app.handle_start_tag(name,attrs)
	else:
	    self.now_at("/>")
	    self.app.handle_start_tag(name,attrs)
	    del self.stack[-1]
	    self.app.handle_end_tag(name)

    def parse_end_tag(self):
	"Parses the end tag from after the '</' and beyond '>'."
	name=self.get_match(reg_name)
	self.skip_ws()
	if not self.now_at(">"): self.err.fatal("'>' expected")

	try:
	    if not name==self.stack[-1]:
		self.err.fatal("End tag for '%s' seen, but '%s' expected" \
			       % (name,self.stack[-1]))

		# Let's do some guessing in case we continue
		if len(self.stack)>1 and self.stack[-2]==name:
		    del self.stack[-1]
		    del self.stack[-1]
	    else:
		del self.stack[-1]
	except IndexError,e:
	    self.err.fatal("Element '%s' not open" % name)

	self.app.handle_end_tag(name)

    def parse_data(self):
	"Parses character data."
	mo=reg_c_data.search(self.data,self.pos)
	if mo==None:
	    if not self.final:
		raise OutOfDataException()

	    start=self.pos
	    end=self.datasize
	    self.pos=self.datasize
	else:
	    start=self.pos
	    end=mo.end(0)-1
	    self.pos=mo.end(0)-1

	if string.find(self.data,"]]>",start,end)!=-1:
	    self.pos=string.find(self.data,"]]>",start,end)
	    self.err.fatal("']]>' must not occur in character data")   

	if self.stack==[]:
	    res=reg_ws.match(self.data,start)

	    if res!=None and res.end(0)==self.datasize:
		dec=1
	    else:
		dec=0

	    if res==None or res.end(0)-dec!=end:
		self.err.fatal("Character data not allowed outside root "
			       "element")
	    
	self.app.handle_data(self.data,start,end)
	
    def parse_charref(self):
	"Parses a character reference."
	if self.now_at("x"):
	    digs=unhex(self.get_match(reg_hex_digits))
	else:
	    digs=string.atoi(self.get_match(reg_digits))

	if not self.now_at(";"): self.err.fatal("';' expected")
	    
	if not (digs==9 or digs==10 or digs==13 or \
		(digs>=32 and digs<=255)):
	    if digs>255:
		self.err.fatal("Unsupported character")
	    else:
		self.err.fatal("Illegal character")
	else:
	    if self.stack==[]:
		self.err.fatal("Character data not allowed outside root "
			       "element")
	    self.app.handle_data(chr(digs),0,1)

    def parse_cdata(self):
	"Parses a CDATA marked section from after the '<![CDATA['."
	new_pos=self.get_index("]]>")
	if self.stack==[]:
	    self.err.fatal("Character data not allowed outside root element")
	self.app.handle_data(self.data,self.pos,new_pos)
	self.pos=new_pos+3

    def parse_ent_ref(self):
	"Parses a general entity reference from after the '&'."
	name=self.get_match(reg_name)
	if not self.now_at(";"): self.err.fatal("';' expected")

        try:
            ent=self.ent.resolve_ge(name)
	except KeyError,e:
	    self.err.fatal("Unknown entity '%s'" % name)
            return

	if ent.name in self.open_ents:
	    self.err.fatal("Entity recursion detected")
	    return
	else:
	    self.open_ents.append(ent.name)

	if self.stack==[]:
	    self.err.fatal("Entity references not allowed outside root "
			   "element")
	    
	if ent.is_internal():
	    self.push_entity(self.get_current_sysid(),ent.value)
	    self.do_parse()
	    self.flush()
	    self.pop_entity()
	else:
	    if ent.notation!="":
		self.err.fatal("Unparsed entities not allowed as general "
			       "entity references in element content")

            sysid=self.app.resolve_entity(ent.get_pubid(),ent.get_sysid())
	    self.open_entity(sysid)

	del self.open_ents[-1]
	
    def parse_doctype(self):
	"Parses the document type declaration."

	if self.seen_doctype:
	    self.err.fatal("Multiple document type declarations")
	if self.seen_root:
	    self.err.fatal("Document type declaration not allowed "
			   "inside root element")
	
	self.skip_ws(1)
	rootname=self.get_match(reg_name)
	self.skip_ws(1)

	if self.now_at("SYSTEM"):
	    self.skip_ws(1)
	    sys_id=self.get_wrapped_match([("\"",reg_sysid_quote),\
					   ("'",reg_sysid_apo)])
	    pub_id=""
	elif self.now_at("PUBLIC"):
	    self.skip_ws(1)
	    pub_id=self.get_wrapped_match([("\"",reg_pubid_quote),\
					   ("'",reg_pubid_apo)])
	    self.skip_ws(1)
	    sys_id=self.get_wrapped_match([("\"",reg_sysid_quote),\
					   ("'",reg_sysid_apo)])
	else:
	    pub_id=""
	    sys_id=""

	self.skip_ws()
	if pub_id!="" or sys_id!="":
	    self.app.handle_doctype(rootname,pub_id,sys_id)
	
	if self.now_at("["):
	    self.parse_internal_dtd()    
	elif not self.now_at(">"):
	    self.err.fatal("'>' expected")

	self.seen_doctype=1 # Has to be at the end to avoid block trouble
    
    def parse_internal_dtd(self):
	"Parse the internal DTD beyond the ']>'."

	self.set_start_point() # Record start of int_subset, preserve data
	self.update_pos()
	line=self.line
	lb=self.last_break
	
	while 1:
	    self.find_reg(reg_int_dtd)

	    if self.now_at("\""): self.scan_to("\"")
	    elif self.now_at("'"): self.scan_to("'")
	    elif self.now_at("<?"): self.scan_to("?>")
	    elif self.now_at("<!--"): self.scan_to("-->")
	    elif self.now_at("<!["): self.scan_to("]]>")
	    elif self.now_at("]>"): break

	# [:-2] cuts of the "]>" at the end
	self.handle_internal_dtd(line,lb,self.get_region()[:-2])

    def parse_xml_decl(self):
	"Parses the contents of the XML declaration from after the '!xml'."

	self.update_pos()
	if self.get_column()!=5 or self.get_line()!=1:
	    self.err.fatal("XML declaration must appear first in "
			   "document")
	if self.seen_xmldecl: # Set in parse_pi, to avoid block problems
	    self.err.fatal("Multiple XML declarations")

	enc=""
	sddecl=""
	self.skip_ws()
	if self.now_at("version"):
	    self.skip_ws()
	    if not self.now_at("="): self.err.fatal("'=' expected")
	    self.skip_ws()
	    ver=self.get_match(reg_ver)[1:-1]
	    if ver!="1.0":
		self.err.fatal("Unsupported XML version")

	    self.skip_ws()
	else:
	    self.err.fatal("XML version info missing on XML declaration")

	try:
	    tst=self.now_at("encoding")
	except OutOfDataException,e:
	    tst=0 # It's OK

	if tst:
	    self.skip_ws()
	    if not self.now_at("="): self.err.fatal("'=' expected")
	    self.skip_ws()
	    enc=self.get_match(reg_enc_name)[1:-1]
	    if enc!="UTF-8" and enc!="ISO-8859-1":
		self.err.fatal("Unsupported character encoding")
	    self.skip_ws()	    

	try:
	    tst=self.now_at("standalone")
	except OutOfDataException,e:
	    tst=0

	if tst:
	    self.skip_ws()
	    if not self.now_at("="): self.err.fatal("'=' expected")
	    self.skip_ws()
	    sddecl=self.get_match(reg_std_alone)[1:-1]
	    self.standalone= sddecl=="yes"

	    self.skip_ws()

	self.skip_ws()
	self.app.set_entity_info(ver,enc,sddecl)
	
    def handle_internal_dtd(self,doctype_line,doctype_lb,int_dtd):
	"Handles the internal DTD."
	p=DTDParser()
	p.set_error_handler(self.err)
	p.set_dtd_consumer(self.dtd)
        if self.dtd_listener!=None:
            self.dtd.set_dtd_listener(self.dtd_listener)
	p.set_internal(1)
	self.err.set_locator(p)

	try:
	    try:		
		p.line=doctype_line
		p.last_break=doctype_lb
		
		p.set_sysid(self.get_current_sysid())
		p.feed(int_dtd)
	    except OutOfDataException,e:
		self.err.fatal("Premature end of internal DTD subset")
	finally:
	    self.err.set_locator(self)
	    self.dtd.dtd_end()
	    
# ==============================
# A DTD parser
# ==============================
	    
class DTDParser(AbstractXML):
    "A parser for XML DTDs, both internal and external."

    def __init__(self):
	EntityParser.__init__(self)
	self.internal=0
	self.dtd=DTDConsumer(self)

	self.ignore=0 # Currently in a conditional section marked ignore?
	self.section_stack=[] # Conditional section nesting tracker

    def set_dtd_consumer(self,dtd):
	"Tells the parser where to send DTD information."
	self.dtd=dtd
	
    def set_internal(self,yesno):
	"Tells the parser whether the DTD is internal or external."
	self.internal=yesno
	
    def do_parse(self):
	"Does the actual parsing."

	try:
	    self.skip_ws()
	    while self.pos<self.datasize:
		prepos=self.pos
		
		if self.now_at("<!ELEMENT"):
		    self.parse_elem_type()
		elif self.now_at("<!ENTITY"):
		    self.parse_entity()
		elif self.now_at("<!ATTLIST"):
		    self.parse_attlist()
		elif self.now_at("<!NOTATION"):
		    self.parse_notation()
		elif self.test_reg(reg_pe_ref):
		    self.parse_pe_ref()
		elif self.now_at("<?"):
		    print "foo"
		    self.parse_pi()
		elif self.now_at("<!--"):
		    self.parse_comment()
		elif self.now_at("<!["):
		    self.parse_conditional()
		elif self.now_at("]]>") and self.section_stack!=[]:
		    self.ignore=self.section_stack[-1]
		    del self.section_stack[-1]
		else:
		    self.err.fatal("Illegal construct")
		    self.pos=self.pos+1

		self.skip_ws()

	except OutOfDataException,e:
	    if self.final:
		raise e
	    else:
		self.pos=prepos

    def parse_entity(self):
	"Parses an entity declaration."

	self.skip_ws(1)
	if self.now_at("%"):
	    pedecl=1
	    self.skip_ws(1)
	else:
	    pedecl=0
	
	ent_name=self.get_match(reg_name)
	self.skip_ws(1)

	pub_id=""
	if self.now_at("SYSTEM"):
	    internal=0
	    self.skip_ws(1)
	    sys_id=self.get_wrapped_match([("\"",reg_sysid_quote),\
					   ("'",reg_sysid_apo)])
	elif self.now_at("PUBLIC"):
	    internal=0
	    self.skip_ws(1)
	    pub_id=self.get_wrapped_match([("\"",reg_pubid_quote),\
					   ("'",reg_pubid_apo)])

	    # Parsing alternate sys_id
	    self.skip_ws(1)
	    sys_id=self.get_wrapped_match([("\"",reg_sysid_quote),\
					   ("'",reg_sysid_apo)])
	else:
	    internal=1
	    ent_val=self.parse_string(0)

	self.skip_ws()
	if not internal and self.now_at("NDATA"):
	    # Parsing the optional NDataDecl
	    if pedecl:
		self.err.fatal("Parameter entities cannot be unparsed")
	    self.skip_ws()

	    ndata=self.get_match(reg_name)
	else:
	    ndata=""
	    
	if not self.now_at(">"):
	    self.err.fatal("Entity declaration incorrectly terminated, '>' "
			   "expected")

        if pedecl:
            if internal:
                self.dtd.new_parameter_entity(ent_name,ent_val)
            else:
                self.dtd.new_external_pe(ent_name,pub_id,sys_id)
        else:
            if internal:
                self.dtd.new_general_entity(ent_name,ent_val)
            else:
                self.dtd.new_external_entity(ent_name,pub_id,sys_id,ndata)

    def parse_notation(self):
	"Parses a notation declaration."
	self.skip_ws(1)
	name=self.get_match(reg_name)
	self.skip_ws(1)

	pubid=""
	sysid=""
	if self.now_at("PUBLIC"):
	    self.skip_ws()
	    pubid=self.get_wrapped_match([("\"",reg_pubid_quote),
					  ("'",reg_pubid_apo)])
	    self.skip_ws(1)
	    if self.test_str("\"") or self.test_str("'"):
		sysid=self.get_wrapped_match([("\"",reg_sysid_quote),
					      ("'",reg_sysid_apo)])
	elif self.now_at("SYSTEM"):
	    self.skip_ws(1)
	    sysid=self.get_wrapped_match([("\"",reg_sysid_quote),
					  ("'",reg_sysid_apo)])
	else:
	    self.err.fatal("'PUBLIC' or 'SYSTEM' expected")
	    self.scan_to(">")
	    self.pos=self.pos-1  # Rewinding to '>'

	self.skip_ws()
	if not self.now_at(">"):
	    self.err.fatal("'>' expected")

	self.dtd.new_notation(name,sysid,pubid)

    def parse_pe_ref(self):
	"Parses a reference to a parameter entity."
	pe_name=self.get_match(reg_pe_ref)[1:-1]

        try:
            ent=self.dtd.resolve_pe(pe_name)
	except KeyError,e:
	    self.err.fatal("Unknown parameter entity '%s'" % name)
            return 

	if ent.is_internal():
	    self.push_entity(self.get_current_sysid(),ent.value)
	    self.do_parse()
	    self.pop_entity()
	else:
	    self.open_entity(ent.sysid) # Does parsing and popping
	    
    def parse_attlist(self):
	"Parses an attribute list declaration."

	self.skip_ws(1)
	elem=self.get_match(reg_name)
	self.skip_ws(1)

	while not self.test_str(">"):
	    attr=self.get_match(reg_name)
	    self.skip_ws(1)

	    if self.test_reg(reg_attr_type):
		a_type=self.get_match(reg_attr_type)
	    elif self.now_at("NOTATION"):
		self.skip_ws(1)
		a_type=("NOTATION",self.__parse_list(reg_name,"|"))
	    elif self.now_at("("):
		self.pos=self.pos-1 # Does not expect '(' to be skipped
		a_type=self.__parse_list(reg_nmtoken,"|")
	    else:
		self.err.fatal("Expected type or alternative list")
		self.scan_to(">")
		return
	    
	    self.skip_ws(1)

	    if self.test_reg(reg_attr_def):
		a_decl=self.get_match(reg_attr_def)
		a_def=""
	    elif self.now_at("#FIXED"):
		self.skip_ws(1)
		a_decl="#FIXED"
		a_def=self.parse_string(0)
	    else:
		a_decl="#DEFAULT"
		a_def=self.parse_string(0)
	    
	    self.skip_ws()

	    self.dtd.new_attribute(elem,attr,a_type,a_decl,a_def)

	self.pos=self.pos+1 # Skipping the '>'

    def parse_elem_type(self):
	"Parses an element type declaration."

	self.skip_ws(1)
	elem_name=self.get_match(reg_name)
	self.skip_ws(1)

	# content-spec
	if self.now_at("EMPTY"):
	    elem_cont=None
	elif self.now_at("ANY"):
	    elem_cont=1 
	elif self.now_at("("):
	    elem_cont=self.parse_content_model()
	else:
	    self.err.fatal("Invalid content declaration")
	    elem_cont=None

	self.skip_ws()
	if not self.now_at(">"):
	    self.err.fatal("Element declaration incorrectly terminated,"+\
			   "'>' expected.")

	self.dtd.new_element_type(elem_name,elem_cont)

    def parse_content_model(self,level=0):
	"""Parses the content model of an element type declaration. Level
	tells the function if we are on the top level (=0) or not (=1)."""

	# Creates a content list with separator first
	cont_list=[]
	sep="" 
	
	if self.now_at("#PCDATA") and level==0:
	    return self.parse_mixed_content_model()

	while 1:
	    self.skip_ws()
	    if self.now_at("("):
		cp=self.parse_content_model(1)
	    else:
		cp=self.get_match(reg_name)

	    self.skip_ws()

	    if self.test_str("?") or self.test_str("*") or self.test_str("+"):
		mod=self.data[self.pos]
		self.pos=self.pos+1
	    else:
		mod=""
	    cont_list.append((cp,mod))

	    self.skip_ws()
	    
	    if self.now_at(")"):
		break

	    if sep=="":
		if self.test_str("|") or self.test_str(","):
		    sep=self.data[self.pos]
		    self.pos=self.pos+1
		else:
		    self.err.fatal("Unknown separator")
	    else:
		if not self.now_at(sep):
		    self.err.fatal("Mixing of choice and sequence lists!")
		    
	if self.test_str("+") or self.test_str("?") or self.test_str("*"):
	    mod=self.data[self.pos]
	    self.pos=self.pos+1
	else:
	    mod=""

	if sep==",":
	    return SeqContentModel(cont_list,mod)
	elif sep=="|":
	    return ChoiceContentModel(cont_list,mod)
	elif sep=="":
	    if mod=="":
		# Pick out modifier from content list, if any
		mod=cont_list[0][1]
	    return ContentModel(cont_list,mod)

    def parse_conditional(self):
	"Parses a conditional section."	
	if self.internal:
	    self.err.fatal("Conditional sections not allowed in internal "
			   "subset")
	    ignore=1
	    self.scan_to("[")
	else:
	    self.skip_ws()

	    if self.now_at("IGNORE"):
		ignore=1
	    elif self.now_at("INCLUDE"):
		ignore=0
	    else:
		self.err.fatal("'IGNORE' or 'INCLUDE' expected")
		self.scan_to("[")
		ignore=1

	    self.skip_ws()
	    if not self.now_at("["):
		self.err.fatal("'[' expected")

	self.section_stack.append(self.ignore)
	self.ignore=ignore or self.ignore		    
	
    def parse_mixed_content_model(self):
	"Parses mixed content models. Ie: ones containing #PCDATA."

	cont_list=[("#PCDATA","")]
	sep="?"

	while 1:
	    self.skip_ws()
	    if self.now_at("|"):
		sep="|"
	    elif self.now_at(")"):
		break
	    else:
		self.err.fatal("'|' expected")

	    self.skip_ws()
	    cont_list.append((self.get_match(reg_name),""))

	if sep=="|" and not self.now_at("*"):
	    self.err.fatal("'*' expected.")

	return ChoiceContentModel(cont_list,"+") 
	
    def __parse_list(self, elem_regexp, separator):
	"Parses a '(' S? elem_regexp S? separator ... ')' list. (Internal.)"

	list=[]
	self.skip_ws()
	if not self.now_at("("):
	    self.err.fatal("'(' expected")

	while 1:
	    self.skip_ws()
	    list.append(self.get_match(elem_regexp))
	    self.skip_ws()
	    if self.now_at(")"):
		break
	    elif not self.now_at(separator):
		self.err.fatal("Expected ')' or '%s'" % separator)
		break

	return list



# basic tests

from xml.parsers import xmllib
import sgmlop
import sys
import time, string

try:
    FILE, VERBOSE = sys.argv[1], 2
except IndexError:
    FILE, VERBOSE = "hamlet.xml", 1

print
print "test collecting parsers on", FILE
print

# --------------------------------------------------------------------
# sgmlop

class myCollector:
    def __init__(self):
	self.data = []
	self.text = []
    def finish_starttag(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("start", tag, data)
    def handle_proc(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("pi", tag, data)
    def handle_special(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("special", data)
    def handle_entityref(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("entity", data)
    def handle_data(self, data):
	self.text.append(data)
    def handle_cdata(self, data):
	self.text.append("CDATA" + data)

t = time.clock()
for i in range(1):
    out = myCollector()
    fp = open(FILE)
    parser = sgmlop.XMLParser()
    parser.register(out)
    b = 0
    while 1:
	data = fp.read(1024)
	if not data:
	    break
	parser.feed(data)
	b = b + len(data)
    parser.close()
t1 = time.clock() - t

print "raw sgmlop", len(out.data), "lines in", round(t1, 3), "seconds;",
print round(b / t1 / 1024, 2), "kbytes per second"

# --------------------------------------------------------------------
# xmllib

class FastXMLParser(xmllib.FastXMLParser):
    def __init__(self):
	xmllib.FastXMLParser.__init__(self)
	self.data = []
	self.text = []
    def unknown_starttag(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("start", tag, data)
    def handle_proc(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("pi", tag, data)
    def handle_special(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("special", data)
    def handle_entityref(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("entity", data)
    def handle_data(self, data):
	self.text.append(data)
    def handle_cdata(self, data):
	self.text.append("CDATA" + data)

t = time.clock()
for i in range(1):
    fp = open(FILE)
    parser2 = FastXMLParser()
    b = 0
    while 1:
	data = fp.read(1024)
	if not data:
	    break
	parser2.feed(data)
	b = b + len(data)
    parser2.close()
t2 = time.clock() - t

print "fast xmllib", len(parser2.data), "lines in", round(t2, 3), "seconds;",
print round(b / t2 / 1024, 2), "kbytes per second"

class SlowXMLParser(xmllib.SlowXMLParser):
    def __init__(self):
	xmllib.SlowXMLParser.__init__(self)
	self.data = []
	self.text = []
    def unknown_starttag(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("start", tag, data)
    def handle_proc(self, tag, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("pi", tag, data)
    def handle_special(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("special", data)
    def handle_entityref(self, data):
	if self.text:
	    self.data.append(repr(string.join(self.text, "")))
	    self.text = []
	self.data.append("entity", data)
    def handle_data(self, data):
	self.text.append(data)
    def handle_cdata(self, data):
	self.text.append("CDATA" + data)

t = time.clock()
for i in range(1):
    fp = open(FILE)
    parser3 = SlowXMLParser()
    b = 0
    while 1:
	data = fp.read(1024)
	if not data:
	    break
	parser3.feed(data)
	b = b + len(data)
    parser3.close()
t3 = time.clock() - t

print "slow xmllib", len(parser3.data), "lines in", round(t3, 3), "seconds;",
print round(b / t3 / 1024, 2), "kbytes per second"

print
print "normalized timing:"
print "slow xmllib", 1.0
print "fast xmllib", round(t2 / t3, 2), "(%sx)" % round(t3 / t2, 1)
print "sgmlop     ", round(t1 / t3, 2), "(%sx)" % round(t3 / t1, 1)
print

print "looking for differences:"

lines = min(len(parser2.data), len(parser3.data))

for i in xrange(lines):
    if parser2.data[i] != parser3.data[i]:
	for j in range(max(i-5, 0), min(i+5, lines)):
	    if parser2.data[j] != parser3.data[j]:
		print "+", j+1, parser2.data[j]
		print "*", j+1, parser3.data[j]
	    else:
		print "=", j+1, parser2.data[j]
	break
else:
    print "   (none)"



import string, re
from xml.dom.core import *

def quote(s):
	s = re.sub('&', '&amp;', s)
	s = re.sub('<', '&lt;', s)
	s = re.sub('>', '&gt;', s)
	return s
	

class XmlLineariser :

	def __init__(self) :
		self.empties = []
		self.strip = []
		self.xml_style_endtags = 1
		self.add_newline_before = []
		self.add_newline_after = []

	def linearise(self, root) :
		if root.NodeType == DOCUMENT:
			assert root.documentElement.NodeType == ELEMENT
			return self.linearise_element(root.documentElement)
		else:
			return self.linearise_element(root)

	def linearise_element(self, element) :
		assert element.NodeType == ELEMENT
		s = ''
		if element.GI in self.add_newline_before:
			s = s + '\n'
		s = s + '<%s' % element.GI
		
		# XXX use DOM interface here.
		for name, value in element.attributes.items() :
			s = s + ' %s="%s"' % (name, value)

		if self.xml_style_endtags and not element.children:
			return s + '/>'
		s = s + '>'

		s1 = ''
		for child in element.children :
			if child.NodeType is ELEMENT:
				s1 = s1 + self.linearise_element(child)
			elif child.NodeType is TEXT:
				s1 = s1 + quote(child.data)
			else :
				s1 = s1 + str(child)

		if element.GI in self.strip :
			s = s + string.strip(s1)
		else :
			s = s + s1
			
		if not element.GI in self.empties :
			s = s + '</%s>' % element.GI
		if element.GI in self.add_newline_after:
			s = s + '\n'

		return s


class HtmlLineariser(XmlLineariser):
	def __init__(self):
		self.xml_style_endtags = 0
		self.empties = [
			'img', 'br', 'hr', 'include', 'li', 'meta', 'input',
			'IMG', 'BR', 'HR', 'INCLUDE', 'LI', 'META', 'INPUT',
		]
		self.strip = [
			'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 
			'li', 'br', 'p', 'a', 'title', 'font',
			'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 
			'LI', 'BR', 'P', 'A', 'TITLE', 'FONT',
		]
		self.add_newline_before = [
			'head', 'body', 'title', 'ul', 'li',
			'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
			'HEAD', 'BODY', 'TITLE', 'UL', 'LI',
			'H1', 'H2', 'H3', 'H4', 'H5', 'H6',
		]
		self.add_newline_after = [
			'html', 'p', 'br', 'hr', 'head', 'body', 'title', 'meta', 'li', 'ul',
			'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
			'HTML', 'P', 'BR', 'HR', 'HEAD', 'BODY', 'TITLE', 'META', 'LI', 'UL',
			'H1', 'H2', 'H3', 'H4', 'H5', 'H6',
		]


class ASPLineariser(XmlLineariser):
	def __init__(self, rep_file):
		self.rep_dict = {}
		self.parseRepFile(rep_file)
		

	def parseRepFile(self, rep_file):
		s = ''
		for l in open(rep_file).readlines():
			if l[0] == '<':
				plus_before = 0
				plus_after = 0
				n = string.index(l, '>')
				tag_name = l[1:n]
				rep = string.strip(l[n+1:])
				if rep and rep[0] == '+':
					plus_before = 1
					rep = string.strip(rep[1:])
				if rep and rep[-1] == '+':
					plus_after = 1
					rep = string.strip(rep[:-1])
				if rep:
					self.rep_dict[tag_name] = (plus_before, plus_after, eval(rep))
				else:
					self.rep_dict[tag_name] = (plus_before, plus_after, '')
					

	def linearise_element(self, element) :
		assert element.NodeType == ELEMENT
		s = ''
		
		# Start tag
		plus_before, plus_after, repl = self.rep_dict[element.getTagName()]

		if s and s[-1] != '\n' and plus_before:
			s = s + '\n'
		s = s + repl
		if s and s[-1] != '\n' and plus_after:
			s = s + '\n'

		s1 = ''
		for child in element.children :
			if child.NodeType is ELEMENT:
				s1 = s1 + self.linearise_element(child)
			elif child.NodeType is TEXT:
				#s1 = s1 + quote(child.data)
				s1 = s1 + child.data
			else :
				s1 = s1 + str(child)
		
		s = s + s1

		# End tag.
		plus_before, plus_after, repl = self.rep_dict['/' + element.getTagName()]
		if s and s[-1] != '\n' and plus_before:
			s = s + '\n'
		s = s + repl
		if s and s[-1] != '\n' and plus_after:
			s = s + '\n'


		return s





from xml.dom.transformer import *
from xml.dom.dc_builder import DcBuilder
from xml.dom.pyhtml import *
from xml.dom.writer import HtmlWriter

test_text = """<?xml>
<doc>
<title>This is a test</title>

<h1>Don't panic</h1>

<p>Maybe it will work.</p>

<h2>We can handle it</>

<h3>Yes we can</h3>
<h3>Or maybe not</h3>

End of test.
</doc>
"""


class TestTransformer(Transformer):
	def do_doc(self, node):
		return [HTML(
			HEAD(
				TITLE(cdata(having(node._children, 'this.GI == "title"')))
			),
			BODY(
				{'bgcolor': '#FFFFFF', 'text': '#000000'},
				having(node._children, 'this.GI in ["h1", "h2", "h3"]')
			)
		)]

transformer = TestTransformer()
parser = DcBuilder()
parser.feed(test_text)
document = transformer.transform(parser.document)

writer = HtmlWriter()
writer.write(document)



"""
A SAX driver for xmlproc with validation and DTD information.
"""

version="0.90"

from xml.sax import saxlib,saxutils
from xml.parsers.xmlproc import xmlval
from xml.sax.drivers.drv_xmlproc import *

import types

# --- SAX_XPValParser

class SAX_XPValParser(SAX_XPParser):

    def __init__(self):
        SAX_XPParser.__init__(self)
    
    def parse(self,sysID):
	self.parser=self.__create_parser()
	self.parser.set_application(self)
        self.parser.set_dtd_listener(self)
	self.parser.set_error_handler(self)
        self.dtd=self.parser.get_dtd()
	self.doc_handler.setDocumentLocator(self)
	self.parser.parse_resource(sysID)
	
    def __create_parser(self):
	return xmlval.XMLValidator()

    def handle_start_tag(self, name, attrs):
        try:
            self.doc_handler.startElement(name,
                                          XPAttributes(attrs,\
                                                       self.dtd.get_elem(name)))
        except KeyError,e:
            self.doc_handler.startElement(name,XPAttributes(attrs,None))
            
    # --- EXPERIMENTAL PYTHON SAX EXTENSIONS:

    def get_parser_name(self):
        return "xmlproc_val"

    def get_driver_version(self):
        return version
    
    def is_validating(self):
        return 1
        
# --- XPAttributes

class XPAttributes(saxutils.AttributeMap):

    def __init__(self,map,elemdecl):
        saxutils.AttributeMap.__init__(self,map)        
        self.elemdecl=elemdecl

        if elemdecl==None:
            self.getType=self.getTypeStatic

    def getTypeStatic(self,i):
        return "CDATA"        # Used for undeclared elements
        
    def getType(self, i):
	if type(i)==types.IntType:
            try:
                i=self.map.keys()[i]
            except KeyError,e:
                return "CDATA"

        try:
            return self.elemdecl.get_attr(i).get_type()
        except KeyError,e:
            return "CDATA"

# --- Global functions

def create_parser():
    return SAX_XPValParser()
    



#!/usr/local/bin/python

"""Extensible Markup Language Scanner, Checker, and Utilities

This xml module includes the followin classes:
  Scanner          -- handles XML syntax only: no well-formedness nor
                      validity constraints
  WellFormed       -- scanner client that
                      well-formedness constraints

  InferEndTags     -- a WellFormed subclass that overrides
                      start tag processing to do a limited
		      form of end-tag inference that is
		      sufficient for HTML 3.2

See: Extensible Markup Language
     http://www.w3.org/TR/WD-xml-lang

This implementation is not quite complete.
There are also some differences from the spec.
@@ marks things I intend to fix. The other differences
are things I probably want to change in the XML spec.
@# marks things that I'm still not comfortable with.

This module also includes a test harness.
 Usage: pyton xml.py [-emit start|end|empty|comment|eref|cref|...]+
                     [-all] < entity


Copyright (c) 1997 by Massachusetts Institute of Technology (MIT),
                      INRIA,
                      Keio
Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
as per <http://www.w3.org/pub/WWW/COPYRIGHT>.

written by Dan Connolly <connolly@w3.org> http://www.w3.org/People/Connolly/
$Id: xml.py,v 1.7 1997/05/22 19:40:10 connolly Exp $
"""

import regex
from string import lower, find, atoi, index, count

#############################################################
# The Client class and supporting utilities, exceptions, etc.
#

Result = 'xml.Done'

Lit        = 'lit'          # quoted literals: "dsj" or 'sdfsd'
Symbol     = 'symbol'       # symbol, e.g. text in <input type=text>
Number     = 'number'       # numbers, e.g. 10 in <img width=10>
                            #@# hmm... float, cardinal, int, units?
Enum       = 'enum'         # enumerated/booleans, e.g. ismap in <img ismap>

SpacePat    = '[ \t\r\n]+'  #@@ unicode space?

class Client:
    def __init__(self):
	pass

    ###########
    # Methods to override to handle parsed data
    #
    # raise Result to tell the scanner to pause
    # and return a result
    #

    def text(self, str):
	pass

    def openStart(self, name):
	pass

    def attribute(self, name, type, value):
	pass

    def closeStart(self):
	pass

    def closeEmpty(self):
	pass

    def endTag(self, name=None):
	pass

    def comment(self, stuff):
	pass

    def pi(self, stuff):
	pass

    def decl(self, name, parts):
	pass

    def cref(self, numeral):
	pass

    def eref(self, name):
	pass

    def eof(self):
	pass


def fixcase(str):
    """Normalize the case of names and symbols."""
    return lower(str)

Entities = {'lt': '<', 'gt': '>', 'amp': '&',
	    'quot': '"', 'apos': "'"}

def evalLit(str, replacements = Entities):
    """Evaluate an XML literal.

    raises index error if ; missing, key error if undefined entity"""

    return unescape(str[1:-1], replacements)

MarkupDelimChars = (('<', '&lt;'), ('>', '&gt;'), ('&', '&amp;'))
LitDelimChars = (('"', '&quot;'), ("'", '&apos;'), ('&', '&amp;'))

def escape(str, replacements = MarkupDelimChars):
    """remember to put & last!"""
    for pat, repl in replacements:
	str = regsub.gsub(pat, repl, str)
    return str

def unescape(str, replacements = Entities):
    """raises index error if ; missing, key error if undefined entity"""
    out = ''
    while 1:
	i = find(str, '&')
	if i < 0:
	    return out + str
	else:
	    out = out + str[:i]
	    str = str[i+1:]
	    j = index(str, ';')
	    out = out + deref(str[:j], replacements)
	    str=str[j+1:]

def deref(str, replacements = Entities):
    """raises index error if ; missing, key error if undefined entity"""
    if str[0] == '#':
	return chr(atoi(str[1:])) #@@ x, hex
    if str[:2] == 'U-':
	return chr(atoi(str[2:], 16))
    else:
	return replacements[str]

def ScanErrorMessage(outfp, info):
    message, line, pending, offending = info
    outfp.write(("ERROR: %d: %s\n" + \
		 "context: %s{!!!}%s\n" ) % \
		(line, message, pending, offending))

def NotWellFormedMessage(outfp, info, wfc):
    message, line, offending = info
    s = wfc.stack
    if s:
	name, ln = s[-1].name, s[-1].line
	ref = "(see <%s> line: %d)" % (name, ln)
    else:
	ref = ''
    outfp.write(("%d: %s %s\n" + \
		 "bad text: %s\n") % \
		(line, message, ref, offending))
    for elm in wfc.stack:
	outfp.write(" line %d: <%s[%d] %s>\n" %
		    (elm.line, elm.name, elm.idx, `elm.attrs`))



ScanError = 'xml.ScanError'

class Scanner:
    ###########
    # Exported Scanner methods

    def __init__(self):
	self._line = 1
	self.done = -1
	self.bb = self.at = 0
	self._buf = ''

        # #@# this scanner doesn't signal an error when ]]> occurs
        # outside a marked section
	self.RE = union(('[^&<]+',
			 '<!--', '<!\[', '<!', '<\?',
			 '</', '<',
			 '&#', '&'))

	self.choices =   (self.noDelim,
			  self.commentDelim, self.sectionDelim,
			    self.declDelim, self.piDelim,
			  self.endDelim, self.startDelim,
			  self.crefDelim, self.erefDelim)

	self.numberRE   = regex.compile('[0-9]+')
	self.spaceRE    = regex.compile(SpacePat)
	self.commentRE  = regex.compile('\(\([^-]\|-[^-]\)*\)-->')
	self.sectionRE  = regex.compile(
	    'CDATA\[\(\([^]]\|\][^]]\|\]\][^>]\)*\)\]\]>')
	self.piRE       = regex.compile('\([^>]*\)>')
	self.litRE      = regex.compile("'[^']*'\|\"[^\"]*\"")
	self.symbolRE   = regex.compile('[a-zA-Z0-9\._-]+')
	self.nameRE     = regex.compile('[a-zA-Z][a-zA-Z0-9\._-]*')


    def client(self, cl):
	#@# multiple clients?
	#@# event masks?
	self.cl = cl

    def feed(self, text):
        self._buf = self._buf + text

    # @# use client=None to refer to the default client which
    #  returns the construct in a canonical form
    def next(self, client):
	while self.at < len(self._buf):
	    i = self.RE.match(self._buf, self.at)
	    if i > 0:
		self.at = self.at + i
		regs = self.RE.regs
		for i in range(0, len(self.choices)):
		    if regs[i+1][1] > 0:
			proc = self.choices[i]
			break

		try:
		    proc(client)
		except Result, r:
		    return r

	    else: raise ScanError, ('illegal delimiter',
				    self._line, '',
				    self._buf[self.at:self.at+25])


    def line(self):
	return self._line

    def raw(self):
	"""get raw text of last construct parsed"""
	return self._buf[self.done:self.bb]

    ###########
    # internal Scanner methods

    def consume(self, more=0):
	at = self.at + more
	self._line = self._line + count(self._buf[:at], '\n', self.bb)
	self.done = self.bb
	self.bb = self.at = at

    def noDelim(self, client):
	self.consume()
	client.text(self.raw())
	
    def name(self, case=0):
        i = self.nameRE.match(self._buf, self.at)
	if i > 0:
	    name = self.nameRE.group(0)
	    self.at = self.at + i
	    if case: name = fixcase(name)
	    return name
	else:
	    return None

    def space(self):
        i = self.spaceRE.match(self._buf, self.at)
	if i > 0:
	    self.at = self.at + i

    def startDelim(self, client):
        name = self.name(1)
	if name:
	    client.openStart(name)

	    while 1:
		self.space()

	        if self._buf[self.at:self.at+2] == '/>':
		    self.consume(2)
		    client.closeEmpty()
		    break

	        elif self._buf[self.at] == '>':
		    self.consume(1)
		    client.closeStart()
		    break
		elif self.at < len(self._buf):
	            self.attribute(client)
	        else:
		    self.err('expected /> or > at end of tag')
	else:
	    self.err('expected tag name after <')

    def attribute(self, client):
	name = self.name()

	if not name:
	    self.err('expected attribute name')

	name = fixcase(name)
        self.space()

	if self._buf[self.at] == '=':
	    self.at = self.at + 1
	    self.space()

	    i = self.litRE.match(self._buf, self.at)
	    if i > 0:
	        lit = self.litRE.group(0)
		self.consume(i)
		client.attribute(name, Lit, lit)
	    #@# else if number...
            else:
		i = self.symbolRE.match(self._buf, self.at)
	        if i > 0:
		    sym = self.symbolRE.group(0)
		    self.consume(i)
		    client.attribute(name, Symbol, sym)
		else:
		    self.err('expected attribute value after =')
        else:
	    self.consume()
	    client.attribute(name, Enum, name)


    def endDelim(self, client):
        name = self.name(1)
	if name:
	    self.space()
	    if(self._buf[self.at] == '>'):
	        self.consume(1)
		client.endTag(name)
	    else:
	        self.err('expected > after end tag name')
        elif self._buf[self.at] == '>':
	    self.consume(1)
	    client.endTag()
        else:
	    self.err('expected end tag name after </')

    def commentDelim(self, client):
        i = self.commentRE.match(self._buf, self.at)
	if i>0:
	    self.consume(i)
	    client.comment(self.commentRE.group(1))
	else:
	    self.err('bad comment')

    def piDelim(self, client):
        i = self.piRE.match(self._buf, self.at)
	if i>0:
	    self.consume(i)
	    client.pi(self.piRE.group(1))
	else:
	    self.err('bad processing instruction')

    def declDelim(self, client):
	name = self.name()
	if not name:
	    self.err('expected declaration keyword after <!')

	name = fixcase(name)

	parts = []

	while 1:
	    self.space()

	    if self._buf[self.at] == '>':
		self.consume(1)
		client.decl(name, parts)
		break
	    else:
		i = self.litRE.match(self._buf, self.at)
		if i > 0:
		    lit = self.litRE.group(0)
		    self.at = self.at + i
		    parts.append((Lit, lit))
		#@# else if number...
		else:
		    i = self.symbolRE.match(self._buf, self.at)
		    if i > 0:
			tok = self.symbolRE.group(0)
			self.at = self.at + i
			parts.append((Symbol, tok))
		    else:
			self.err('bad declaration')

    def sectionDelim(self, client):
	#@# add support for <![xyz[ ... ]xyz]>
        i = self.sectionRE.match(self._buf, self.at)
	if i>0:
	    self.consume(i)
	    client.text(self.commentRE.group(1))
	else:
	    self.err('bad marked section')


    def crefDelim(self, client):
	# @@ unicode &$x
        i = self.numberRE.match(self._buf, self.at)
	if i > 0:
	    num = self.numberRe.group(0)
	    self.at = self.at + i
	    if self._buf[self.at + i] == ';':
	        self.consume(1)
	        client.cref(num)
	    else:
	        self.err('expected ; after character number')
        else:
	    self.err('expected character number after &#')

    def erefDelim(self, client):
        name = self.name()
        if name:
	    if self._buf[self.at] == ';':
	        self.consume(1)
	        client.eref(name)
	    else:
	        self.err('expected ; after entity name')
        else:
	    self.err('expected entity name after &')

    def err(self, msg):
    	raise ScanError, (msg,
			  self._line,
			  self._buf[self.bb:self.at],
			  self._buf[self.at:self.at+25])


def union(choices):
    exp = ''
    for sub in choices:
	if exp: exp = exp + '\|'
	exp = exp + '\(' + sub + '\)'
    return regex.compile(exp)



NotWellFormed = 'xml.NotWellFormed'

#@@ split client interface in two. Clean up empty=start/end

class WellFormed(Client):
    """XML scanner client that enforces well-formedness constraints

    Also maintains a stack of open elements."""

    def __init__(self):
	Client.__init__(self)
        self.stack = []
	self.element = 1
	self.elm = None
	self.spaceRE = regex.compile(SpacePat)

    def startElement(self, elm):
	pass

    def endElement(self, elm):
	pass

    def scanner(self, s):
	"""Tell self which scanner to call to get line numbers and raw input."""
	self.p = s

    def text(self, str):
	# @# hmmm... I think this should be a VC, not a WFC
	if not self.stack:
	    if self.spaceRE.match(str) != len(str):
		raise NotWellFormed, ('data outside root element',
				      self.p.line(), self.p.raw())

    def cref(self, numeral):
	self.text('#' + deref(numeral))

    def eref(self, name):
	try:
	    t = deref(name)
	except KeyError:
	    #@@ Don't silently fix errors!
	    self.text(self.p.raw())
	    return
	self.text(t)

    def openStart(self, name):
	self.elm = Element(name, self.element, self.p.line())

    def attribute(self, name, type, value):
	if self.elm.attrs.has_key(name):
	    raise NotWellFormed, ('duplicate attribute name',
				  self.p.line(), name)
	if type == Lit: self.elm.attrs[name] = evalLit(value)
	else: self.elm.attrs[name] = value

    def closeStart(self):
	self.stack.append(self.elm)
	self.startElement(self.elm)
	self.element = self.element + 1

    def closeEmpty(self):
	self.stack.append(self.elm)
	self.startElement(self.elm)
	self.endElement(self.elm)
	del self.stack[-1]

    def endTag(self, name = None):
	if len(self.stack) == 0:
	    raise NotWellFormed, ('extra end tag',
				  self.p.line(), self.p.raw())

	elm = self.stack[-1]
	if name is None or name == elm.name:
	    self.endElement(elm)
	    del self.stack[-1]
	else:
	    l = map(lambda x : x.name, self.stack)
	    raise NotWellFormed, ('end tag mismatch',
				  self.p.line(), self.p.raw(), l)

    #@# check declarations. Hmmm... VC or WFC?

    def eof(self):
	if len(self.stack) > 0:
	    raise NotWellFormed, ('not enough end tags', self.p.line(), '')


class Element:
    def __init__(self, name, idx, line):
	self.name, self.line = name, line
	self.attrs = {}
	self.idx = idx



class InferEndTags(WellFormed):
    """This subclass of WellFormed does rudimentary end-tag inference.
    It can process empty elements, given a list of their names,
    and given a mapping from elements to their "follow" set
    (i.e. the set of elements which cannot occur in content), it does
    end-tag inference.

    @#It works for HTML 3.2, but I haven't considered the general case
    carefully. I wonder if there is a class of documents including
    HTML 3.2 which is smaller than the general case but still interesting.
    """

    def __init__(self):
	WellFormed.__init__(self)
	self.inferEnds = {}
	self.empties = []

    # @# should have openStart() method with Element subclass

    def closeStart(self):
	elm = self.elm
	name = elm.name
	while self.stack:
	    if name in self.stack[-1].follow:
		self.endTag()
	    else:
		break

	if self.inferEnds.has_key(name):
	    elm.follow = self.inferEnds[name]
	else:
	    elm.follow = []

	if elm.name in self.empties:
	    WellFormed.closeEmpty(self)
	else:
	    WellFormed.closeStart(self)

    def endTag(self, name = None):
	if name:
	    while self.stack:
		top = self.stack[-1]
		if top.follow and top.name != name:
		    WellFormed.endTag(self, None)
		else:
		    break

	WellFormed.endTag(self, name)

    def html32(self):
	self.empties = [
	    'link', 'meta', 'base',
	    'br', 'hr',
	    'img', 'param',
	    'input', 'option', 'isindex'
	    ]
	list = ('ol', 'ul', 'dl')
	heading = ('h1', 'h2', 'h3', 'h4', 'h5', 'h6')
	blocks = ('p', 'address', 'blockquote', 'form', 'table', 'pre') + \
		 heading + list
	self.inferEnds = {
	    'p': blocks,
	    
	    'li': ('li',),
	    'dt': ('dt',),
	    'dd': ('dt', 'dd'),

	    'tr': ('tr',),
	    'th': ('th', 'td', 'tr'),
	    'td': ('th', 'td', 'tr'),

	    }



###############
# Test harness
#

Text       = 'text'         # text data characters
Start      = 'start'        # start element
CloseStart = 'closestart'   # start tag close
CloseEmpty = 'empty'        # empty tag
End        = 'end'          # end element
EndTag     = 'endtag'       # end element
ERef       = 'eref'         # entity reference: &foo;
CRef       = 'cref'         # character reference: &#100;
Comment    = 'comment'      # <!-- bla bla -->
Section    = 'section'      # <![CDATA[ bla bla ]]>
Decl       = 'decl'         # <!doctype ...>
PI         = 'pi'           # <?...>

Events = {Text:Text,
	  Start:Start, End:End,
	  CloseStart:CloseStart, EndTag:EndTag, CloseEmpty: CloseEmpty, 
	  ERef:ERef, CRef:CRef,
	  Comment:Comment, Decl:Decl, PI:PI, Section:Section}

class Tester(InferEndTags):
    def __init__(self):
	InferEndTags.__init__(self)
	self.mask = []

    def showEvents(self, mask):
	self.mask = mask

    def process(self, type, info):
	text = self.p.raw()
	if self.mask is None or type in self.mask:
	    print "==>%s: %s [%s]\n" % (type, `info`, text)

    def text(self, text):
	InferEndTags.text(self, text)
	self.process(Text, None)

    def closeStart(self):
	InferEndTags.closeStart(self)
	elm = self.elm
	self.process(CloseStart, (elm.name, elm.idx, elm.line, elm.attrs))

    def startElement(self, elm):
	names = []
	for e in self.stack:
	    names.append(e.name)

	self.process(Start, (names, elm.idx, elm.line, elm.attrs))

    def closeEmpty(self):
	InferEndTags.closeEmpty(self)
	elm = self.elm
	self.process(CloseEmpty, (elm.name, elm.idx, elm.line, elm.attrs))

    def endTag(self, name = None):
	InferEndTags.endTag(self, name)
	self.process(EndTag, name)

    def endElement(self, elm):
	self.process(End, elm.name)

    def cref(self, num):
	InferEndTags.cref(self, num)
	self.process(CRef, num)

    def eref(self, num):
	InferEndTags.eref(self, num)
	self.process(ERef, num)

    def comment(self, str):
	InferEndTags.comment(self, str)
	self.process(Comment, str)

    def pi(self, str):
	InferEndTags.pi(self, str)
	self.process(PI, str)

    def decl(self, name, parts):
	InferEndTags.decl(self, name, parts)
	self.process(Decl, (name, parts))

    def section(self):
	InferEndTags.section(self, str)
	self.process(Section, None)


def main(infp):
    p = Scanner()
    cl = Tester()
    cl.scanner(p)
    getOpts(cl)

    try:
        p.feed(infp.read())
	p.next(cl)
	cl.eof()

    except ScanError, info:
	ScanErrorMessage(sys.stderr, info)
	sys.exit(1)
    except NotWellFormed, info:
	NotWellFormedMessage(sys.stderr, info, cl)
	sys.exit(1)


def getOpts(x):
    mask = []

    while len(sys.argv) > 1:
        opt = sys.argv[1]
	if opt[0] != '-': break
	del sys.argv[1]

	if opt == '-emit':
	    mask.append(Events[sys.argv[1]])
	    del sys.argv[1]

	elif opt == '-all':
	    mask = None
	elif opt == '-html':
	    x.html32()

	else:
	    sys.stderr.write("unkonwn option: %s\n" % opt)
	    sys.exit(1)

    x.showEvents(mask)


if __name__ == '__main__':
    import sys
    main(sys.stdin)

# vim:ts=8
